import tensorflow as tf
from tensorflow.python import ops, math_ops, state_ops, control_flow_ops
from tensorflow.python.keras.optimizer_v2.optimizer_v2 import OptimizerV2

from notekeras.backend import backend

__all__ = ['AdamWarmup']


class AdamWarmup(OptimizerV2):
    """Adam optimizer with warmup."""

    def __init__(self,
                 decay_steps,
                 warmup_steps,
                 min_lr=0.0,
                 learning_rate=0.001,
                 beta_1=0.9,
                 beta_2=0.999,
                 epsilon=1e-7,
                 weight_decay=0.,
                 weight_decay_pattern=None,
                 amsgrad=False,
                 name='Adam',
                 **kwargs):
        r"""Construct a new Adam optimizer.

        Args:
            decay_steps: Learning rate will decay linearly to zero in decay steps.
            warmup_steps: Learning rate will increase linearly to lr in first warmup steps.
            lr: float >= 0. Learning rate.
            beta_1: float, 0 < beta < 1. Generally close to 1.
            beta_2: float, 0 < beta < 1. Generally close to 1.
            epsilon: float >= 0. Fuzz factor. If `None`, defaults to `K.epsilon()`.
            weight_decay: float >= 0. Weight decay.
            weight_decay_pattern: A list of strings. The substring of weight names to be decayed.
                                  All weights will be decayed if it is None.
            amsgrad: boolean. Whether to apply the AMSGrad variant of this
                algorithm from the paper "On the Convergence of Adam and
                Beyond".
        """

        super(AdamWarmup, self).__init__(name, **kwargs)
        self._set_hyper('decay_steps', float(decay_steps))
        self._set_hyper('warmup_steps', float(warmup_steps))
        self._set_hyper('min_lr', min_lr)
        self._set_hyper('learning_rate', kwargs.get('lr', learning_rate))
        self._set_hyper('decay', self._initial_decay)
        self._set_hyper('beta_1', beta_1)
        self._set_hyper('beta_2', beta_2)
        self._set_hyper('weight_decay', weight_decay)
        self.epsilon = epsilon or backend.epsilon()
        self.amsgrad = amsgrad
        self._initial_weight_decay = weight_decay
        self._weight_decay_pattern = weight_decay_pattern

    def _create_slots(self, var_list):
        for var in var_list:
            self.add_slot(var, 'm')
        for var in var_list:
            self.add_slot(var, 'v')
        if self.amsgrad:
            for var in var_list:
                self.add_slot(var, 'vhat')

    def set_weights(self, weights):
        params = self.weights
        num_vars = int((len(params) - 1) / 2)
        if len(weights) == 3 * num_vars + 1:
            weights = weights[:len(params)]
        super(AdamWarmup, self).set_weights(weights)

    def _resource_apply_dense(self, grad, var):
        var_dtype = var.dtype.base_dtype
        lr_t = self._decayed_lr(var_dtype)
        m = self.get_slot(var, 'm')
        v = self.get_slot(var, 'v')
        beta_1_t = self._get_hyper('beta_1', var_dtype)
        beta_2_t = self._get_hyper('beta_2', var_dtype)
        epsilon_t = ops.convert_to_tensor(self.epsilon, var_dtype)
        local_step = math_ops.cast(self.iterations + 1, var_dtype)
        beta_1_power = math_ops.pow(beta_1_t, local_step)
        beta_2_power = math_ops.pow(beta_2_t, local_step)

        decay_steps = self._get_hyper('decay_steps', var_dtype)
        warmup_steps = self._get_hyper('warmup_steps', var_dtype)
        min_lr = self._get_hyper('min_lr', var_dtype)
        lr_t = tf.where(
            local_step <= warmup_steps,
            lr_t * (local_step / warmup_steps),
            min_lr + (lr_t - min_lr) * (1.0 - tf.minimum(local_step, decay_steps) / decay_steps),
        )
        lr_t = (lr_t * math_ops.sqrt(1 - beta_2_power) / (1 - beta_1_power))

        m_t = state_ops.assign(m,
                               beta_1_t * m + (1.0 - beta_1_t) * grad,
                               use_locking=self._use_locking)

        v_t = state_ops.assign(v,
                               beta_2_t * v + (1.0 - beta_2_t) * math_ops.square(grad),
                               use_locking=self._use_locking)

        if self.amsgrad:
            v_hat = self.get_slot(var, 'vhat')
            v_hat_t = math_ops.maximum(v_hat, v_t)
            var_update = m_t / (math_ops.sqrt(v_hat_t) + epsilon_t)
        else:
            var_update = m_t / (math_ops.sqrt(v_t) + epsilon_t)

        if self._initial_weight_decay > 0.0:
            weight_decay = self._get_hyper('weight_decay', var_dtype)
            var_update += weight_decay * var
        var_update = state_ops.assign_sub(var, lr_t * var_update, use_locking=self._use_locking)

        updates = [var_update, m_t, v_t]
        if self.amsgrad:
            updates.append(v_hat_t)
        return control_flow_ops.group(*updates)

    def _resource_apply_sparse(self, grad, var, indices):
        var_dtype = var.dtype.base_dtype
        lr_t = self._decayed_lr(var_dtype)
        beta_1_t = self._get_hyper('beta_1', var_dtype)
        beta_2_t = self._get_hyper('beta_2', var_dtype)
        epsilon_t = ops.convert_to_tensor(self.epsilon, var_dtype)
        local_step = math_ops.cast(self.iterations + 1, var_dtype)
        beta_1_power = math_ops.pow(beta_1_t, local_step)
        beta_2_power = math_ops.pow(beta_2_t, local_step)

        decay_steps = self._get_hyper('decay_steps', var_dtype)
        warmup_steps = self._get_hyper('warmup_steps', var_dtype)
        min_lr = self._get_hyper('min_lr', var_dtype)
        lr_t = tf.where(
            local_step <= warmup_steps,
            lr_t * (local_step / warmup_steps),
            min_lr + (lr_t - min_lr) * (1.0 - tf.minimum(local_step, decay_steps) / decay_steps),
        )
        lr_t = (lr_t * math_ops.sqrt(1 - beta_2_power) / (1 - beta_1_power))

        m = self.get_slot(var, 'm')
        m_scaled_g_values = grad * (1 - beta_1_t)
        m_t = state_ops.assign(m, m * beta_1_t, use_locking=self._use_locking)
        with ops.control_dependencies([m_t]):
            m_t = self._resource_scatter_add(m, indices, m_scaled_g_values)

        v = self.get_slot(var, 'v')
        v_scaled_g_values = (grad * grad) * (1 - beta_2_t)
        v_t = state_ops.assign(v, v * beta_2_t, use_locking=self._use_locking)
        with ops.control_dependencies([v_t]):
            v_t = self._resource_scatter_add(v, indices, v_scaled_g_values)

        if self.amsgrad:
            v_hat = self.get_slot(var, 'vhat')
            v_hat_t = math_ops.maximum(v_hat, v_t)
            var_update = m_t / (math_ops.sqrt(v_hat_t) + epsilon_t)
        else:
            var_update = m_t / (math_ops.sqrt(v_t) + epsilon_t)

        if self._initial_weight_decay > 0.0:
            weight_decay = self._get_hyper('weight_decay', var_dtype)
            var_update += weight_decay * var
        var_update = state_ops.assign_sub(var, lr_t * var_update, use_locking=self._use_locking)

        updates = [var_update, m_t, v_t]
        if self.amsgrad:
            updates.append(v_hat_t)
        return control_flow_ops.group(*updates)

    def get_config(self):
        config = super(AdamWarmup, self).get_config
        config.update({
            'decay_steps': self._serialize_hyperparameter('decay_steps'),
            'warmup_steps': self._serialize_hyperparameter('warmup_steps'),
            'min_lr': self._serialize_hyperparameter('min_lr'),
            'learning_rate': self._serialize_hyperparameter('learning_rate'),
            'decay': self._serialize_hyperparameter('decay'),
            'beta_1': self._serialize_hyperparameter('beta_1'),
            'beta_2': self._serialize_hyperparameter('beta_2'),
            'weight_decay': self._serialize_hyperparameter('weight_decay'),
            'epsilon': self.epsilon,
            'amsgrad': self.amsgrad,
        })
        return config



import os
from collections import namedtuple
from functools import reduce

import numpy as np
from matplotlib.colors import hsv_to_rgb, rgb_to_hsv
from notekeras.backend import backend as K
from notekeras.backend import keras
from notekeras.layers import Extract, MaskedGlobalMaxPool1D
from notekeras.models.loader import (load_trained_model_from_checkpoint,
                                     load_vocabulary)
from notekeras.tokenizer import Tokenizer
from PIL import Image

__all__ = [
    'POOL_NSP', 'POOL_MAX', 'POOL_AVE',
    'get_checkpoint_paths', 'extract_embeddings_generator', 'extract_embeddings', 'compose', 'get_random_data'
]

POOL_NSP = 'POOL_NSP'
POOL_MAX = 'POOL_MAX'
POOL_AVE = 'POOL_AVE'


def letterbox_image(image, size):
    '''resize image with unchanged aspect ratio using padding'''
    iw, ih = image.size
    w, h = size
    scale = min(w / iw, h / ih)
    nw = int(iw * scale)
    nh = int(ih * scale)

    image = image.resize((nw, nh), Image.BICUBIC)
    new_image = Image.new('RGB', size, (128, 128, 128))
    new_image.paste(image, ((w - nw) // 2, (h - nh) // 2))
    return new_image


def rand(a=0, b=1):
    return np.random.rand() * (b - a) + a


def get_random_data(annotation_line,
                    input_shape,
                    random=True,
                    max_boxes=20,
                    jitter=.3,
                    hue=.1,
                    sat=1.5,
                    val=1.5,
                    proc_img=True):
    '''random preprocessing for real-time data augmentation'''
    line = annotation_line.split()
    image = Image.open(line[0])
    iw, ih = image.size
    h, w = input_shape
    box = np.array([np.array(list(map(int, box.split(',')))) for box in line[1:]])

    if not random:
        # resize image
        scale = min(w / iw, h / ih)
        nw = int(iw * scale)
        nh = int(ih * scale)
        dx = (w - nw) // 2
        dy = (h - nh) // 2
        image_data = 0
        if proc_img:
            image = image.resize((nw, nh), Image.BICUBIC)
            new_image = Image.new('RGB', (w, h), (128, 128, 128))
            new_image.paste(image, (dx, dy))
            image_data = np.array(new_image) / 255.

        # correct boxes
        box_data = np.zeros((max_boxes, 5))
        if len(box) > 0:
            np.random.shuffle(box)
            if len(box) > max_boxes: box = box[:max_boxes]
            box[:, [0, 2]] = box[:, [0, 2]] * scale + dx
            box[:, [1, 3]] = box[:, [1, 3]] * scale + dy
            box_data[:len(box)] = box

        return image_data, box_data

    # resize image
    new_ar = w / h * rand(1 - jitter, 1 + jitter) / rand(1 - jitter, 1 + jitter)
    scale = rand(.25, 2)
    if new_ar < 1:
        nh = int(scale * h)
        nw = int(nh * new_ar)
    else:
        nw = int(scale * w)
        nh = int(nw / new_ar)
    image = image.resize((nw, nh), Image.BICUBIC)

    # place image
    dx = int(rand(0, w - nw))
    dy = int(rand(0, h - nh))
    new_image = Image.new('RGB', (w, h), (128, 128, 128))
    new_image.paste(image, (dx, dy))
    image = new_image

    # flip image or not
    flip = rand() < .5
    if flip: image = image.transpose(Image.FLIP_LEFT_RIGHT)

    # distort image
    hue = rand(-hue, hue)
    sat = rand(1, sat) if rand() < .5 else 1 / rand(1, sat)
    val = rand(1, val) if rand() < .5 else 1 / rand(1, val)
    x = rgb_to_hsv(np.array(image) / 255.)
    x[..., 0] += hue
    x[..., 0][x[..., 0] > 1] -= 1
    x[..., 0][x[..., 0] < 0] += 1
    x[..., 1] *= sat
    x[..., 2] *= val
    x[x > 1] = 1
    x[x < 0] = 0
    image_data = hsv_to_rgb(x)  # numpy array, 0 to 1

    # correct boxes
    box_data = np.zeros((max_boxes, 5))
    if len(box) > 0:
        np.random.shuffle(box)
        box[:, [0, 2]] = box[:, [0, 2]] * nw / iw + dx
        box[:, [1, 3]] = box[:, [1, 3]] * nh / ih + dy
        if flip: box[:, [0, 2]] = w - box[:, [2, 0]]
        box[:, 0:2][box[:, 0:2] < 0] = 0
        box[:, 2][box[:, 2] > w] = w
        box[:, 3][box[:, 3] > h] = h
        box_w = box[:, 2] - box[:, 0]
        box_h = box[:, 3] - box[:, 1]
        box = box[np.logical_and(box_w > 1, box_h > 1)]  # discard invalid box
        if len(box) > max_boxes: box = box[:max_boxes]
        box_data[:len(box)] = box

    return image_data, box_data


def compose(*funcs):
    """Compose arbitrarily many functions, evaluated left to right.
    函数嵌套 compose(f,g,k) = k(g(f))
    Reference: https://mathieularose.com/function-composition-in-python/
    """
    # return lambda x: reduce(lambda v, f: f(v), funcs, x)
    if funcs:
        return reduce(lambda f, g: lambda *a, **kw: g(f(*a, **kw)), funcs)
    else:
        raise ValueError('Composition of empty sequence not supported.')


def get_checkpoint_paths(model_path):
    CheckpointPaths = namedtuple('CheckpointPaths', ['config', 'checkpoint', 'vocab'])
    config_path = os.path.join(model_path, 'bert_config.json')
    checkpoint_path = os.path.join(model_path, 'bert_model.ckpt')
    vocab_path = os.path.join(model_path, 'vocab.txt')
    return CheckpointPaths(config_path, checkpoint_path, vocab_path)


def extract_embeddings_generator(model,
                                 texts,
                                 poolings=None,
                                 vocabs=None,
                                 cased=False,
                                 batch_size=4,
                                 cut_embed=True,
                                 output_layer_num=1):
    """Extract embeddings from texts.

    :param model: Path to the checkpoint or built model without MLM and NSP.
    :param texts: Iterable texts.
    :param poolings: Pooling methods. Word embeddings will be returned if it is None.
                     Otherwise concatenated pooled embeddings will be returned.
    :param vocabs: A dict should be provided if model is built.
    :param cased: Whether it is cased for tokenizer.
    :param batch_size: Batch size.
    :param cut_embed: The computed embeddings will be cut based on their input lengths.
    :param output_layer_num: The number of layers whose outputs will be concatenated as a single output.
                             Only available when `model` is a path to checkpoint.
    :return: A list of numpy arrays representing the embeddings.
    """
    if isinstance(model, (str, type(u''))):
        paths = get_checkpoint_paths(model)
        model = load_trained_model_from_checkpoint(
            config_file=paths.config,
            checkpoint_file=paths.checkpoint,
            output_layer_num=output_layer_num,
        )
        vocabs = load_vocabulary(paths.vocab)

    seq_len = K.int_shape(model.outputs[0])[1]
    tokenizer = Tokenizer(vocabs, cased=cased)

    def _batch_generator():
        tokens, segments = [], []

        def _pad_inputs():
            if seq_len is None:
                max_len = max(map(len, tokens))
                for i in range(len(tokens)):
                    tokens[i].extend([0] * (max_len - len(tokens[i])))
                    segments[i].extend([0] * (max_len - len(segments[i])))
            return [np.array(tokens), np.array(segments)]

        for text in texts:
            if isinstance(text, (str, type(u''))):
                token, segment = tokenizer.encode(text, max_len=seq_len)
            else:
                token, segment = tokenizer.encode(text[0], text[1], max_len=seq_len)
            tokens.append(token)
            segments.append(segment)
            if len(tokens) == batch_size:
                yield _pad_inputs()
                tokens, segments = [], []
        if len(tokens) > 0:
            yield _pad_inputs()

    if poolings is not None:
        if isinstance(poolings, (str, type(u''))):
            poolings = [poolings]
        outputs = []
        for pooling in poolings:
            if pooling == POOL_NSP:
                outputs.append(Extract(index=0, name='Pool-NSP')(model.outputs[0]))
            elif pooling == POOL_MAX:
                outputs.append(MaskedGlobalMaxPool1D(name='Pool-Max')(model.outputs[0]))
            elif pooling == POOL_AVE:
                outputs.append(keras.layers.GlobalAvgPool1D(name='Pool-Ave')(model.outputs[0]))
            else:
                raise ValueError('Unknown pooling method: {}'.format(pooling))
        if len(outputs) == 1:
            outputs = outputs[0]
        else:
            outputs = keras.layers.Concatenate(name='Concatenate')(outputs)
        model = keras.models.Model(inputs=model.inputs, outputs=outputs)

    for batch_inputs in _batch_generator():
        outputs = model.predict(batch_inputs)
        for inputs, output in zip(batch_inputs[0], outputs):
            if poolings is None and cut_embed:
                length = 0
                for i in range(len(inputs) - 1, -1, -1):
                    if inputs[i] != 0:
                        length = i + 1
                        break
                output = output[:length]
            yield output


def extract_embeddings(model,
                       texts,
                       poolings=None,
                       vocabs=None,
                       cased=False,
                       batch_size=4,
                       cut_embed=True,
                       output_layer_num=1):
    """Extract embeddings from texts.

    :param model: Path to the checkpoint or built model without MLM and NSP.
    :param texts: Iterable texts.
    :param poolings: Pooling methods. Word embeddings will be returned if it is None.
                     Otherwise concatenated pooled embeddings will be returned.
    :param vocabs: A dict should be provided if model is built.
    :param cased: Whether it is cased for tokenizer.
    :param batch_size: Batch size.
    :param cut_embed: The computed embeddings will be cut based on their input lengths.
    :param output_layer_num: The number of layers whose outputs will be concatenated as a single output.
                             Only available when `model` is a path to checkpoint.
    :return: A list of numpy arrays representing the embeddings.
    """
    return [embedding for embedding in extract_embeddings_generator(
        model, texts, poolings, vocabs, cased, batch_size, cut_embed, output_layer_num
    )]



import tensorflow as tf
from tensorflow.keras.layers import BatchNormalization, Layer


class Dice(Layer):
    def __init__(self):
        super(Dice, self).__init__()
        self.bn = None
        self.alpha = None

    def build(self, input_shape):
        self.bn = BatchNormalization(center=False, scale=False)
        self.alpha = self.add_weight(shape=(), dtype=tf.float32, name='alpha')

    def call(self, inputs, **kwargs):
        inputs_normed = self.bn(inputs)
        inputs_p = tf.sigmoid(inputs_normed)

        return self.alpha * (1.0 - inputs_p) * inputs + inputs_p * inputs



import colorsys
import random

import cv2
import numpy as np
import tensorflow as tf

# CLASSES = "/root/workspace/notechats/notekeras/example/yolo4/data/classes/coco.names"

ANCHORS = [12, 16, 19, 36, 40, 28, 36, 75, 76,
           55, 72, 146, 142, 110, 192, 243, 459, 401]
ANCHORS_V3 = [10, 13, 16, 30, 33, 23, 30, 61,
              62, 45, 59, 119, 116, 90, 156, 198, 373, 326]
ANCHORS_TINY = [23, 27, 37, 58, 81, 82, 81, 82, 135, 169, 344, 319]
STRIDES = [8, 16, 32]
STRIDES_TINY = [16, 32]
XYSCALE = [1.2, 1.1, 1.05]
XYSCALE_TINY = [1.05, 1.05]
ANCHOR_PER_SCALE = 3
IOU_LOSS_THRESH = 0.5


def load_freeze_layer(model='yolov4', tiny=False):
    if tiny:
        if model == 'yolov3':
            freeze_layouts = ['conv2d_9', 'conv2d_12']
        else:
            freeze_layouts = ['conv2d_17', 'conv2d_20']
    else:
        if model == 'yolov3':
            freeze_layouts = ['conv2d_58', 'conv2d_66', 'conv2d_74']
        else:
            freeze_layouts = ['conv2d_93', 'conv2d_101', 'conv2d_109']
    return freeze_layouts


def load_weights(model, weights_file, model_name='yolov4'):
    if model_name == 'yolov3':
        layer_size = 75
        output_pos = [58, 66, 74]
    else:
        layer_size = 110
        output_pos = [93, 101, 109]

    wf = open(weights_file, 'rb')
    major, minor, revision, seen, _ = np.fromfile(wf, dtype=np.int32, count=5)

    j = 0
    for i in range(layer_size):
        conv_layer_name = 'conv2d_%d' % i if i > 0 else 'conv2d'
        bn_layer_name = 'batch_normalization_%d' % j if j > 0 else 'batch_normalization'

        conv_layer = model.get_layer(conv_layer_name)
        filters = conv_layer.filters
        k_size = conv_layer.kernel_size[0]
        in_dim = conv_layer.input_shape[-1]

        if i not in output_pos:
            # darknet weights: [beta, gamma, mean, variance]
            bn_weights = np.fromfile(wf, dtype=np.float32, count=4 * filters)
            # tf weights: [gamma, beta, mean, variance]
            bn_weights = bn_weights.reshape((4, filters))[[1, 0, 2, 3]]
            bn_layer = model.get_layer(bn_layer_name)
            j += 1
        else:
            conv_bias = np.fromfile(wf, dtype=np.float32, count=filters)

        # darknet shape (out_dim, in_dim, height, width)
        conv_shape = (filters, in_dim, k_size, k_size)
        conv_weights = np.fromfile(
            wf, dtype=np.float32, count=np.product(conv_shape))
        # tf shape (height, width, in_dim, out_dim)
        conv_weights = conv_weights.reshape(conv_shape).transpose([2, 3, 1, 0])

        if i not in output_pos:
            conv_layer.set_weights([conv_weights])
            bn_layer.set_weights(bn_weights)
        else:
            conv_layer.set_weights([conv_weights, conv_bias])

    # assert len(wf.read()) == 0, 'failed to read all data'
    wf.close()


def read_class_names(class_file_name):
    names = {}
    with open(class_file_name, 'r') as data:
        for ID, name in enumerate(data):
            names[ID] = name.strip('\n')
    return names


def load_config(model='yolov4', num_class=80):
    _STRIDES = np.array(STRIDES)
    if model == 'yolov4':
        _ANCHORS = get_anchors(ANCHORS)
    elif model == 'yolov3':
        _ANCHORS = get_anchors(ANCHORS_V3)
    _XYSCALE = XYSCALE if model == 'yolov4' else [1, 1, 1]

    return _STRIDES, _ANCHORS, num_class, _XYSCALE


def get_anchors(anchors_path):
    anchors = np.array(anchors_path)
    return anchors.reshape(3, 3, 2)


def image_preprocess(image, target_size, gt_boxes=None):
    ih, iw = target_size
    h, w, _ = image.shape

    scale = min(iw / w, ih / h)
    nw, nh = int(scale * w), int(scale * h)
    image_resized = cv2.resize(image, (nw, nh))

    image_paded = np.full(shape=[ih, iw, 3], fill_value=128.0)
    dw, dh = (iw - nw) // 2, (ih - nh) // 2
    image_paded[dh:nh + dh, dw:nw + dw, :] = image_resized
    image_paded = image_paded / 255.

    if gt_boxes is None:
        return image_paded

    else:
        gt_boxes[:, [0, 2]] = gt_boxes[:, [0, 2]] * scale + dw
        gt_boxes[:, [1, 3]] = gt_boxes[:, [1, 3]] * scale + dh
        return image_paded, gt_boxes


def draw_bbox(image, bboxes, classes, show_label=True):
    num_classes = len(classes)
    image_h, image_w, _ = image.shape
    hsv_tuples = [(1.0 * x / num_classes, 1., 1.) for x in range(num_classes)]
    colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
    colors = list(
        map(lambda x: (int(x[0] * 255), int(x[1] * 255), int(x[2] * 255)), colors))

    random.seed(0)
    random.shuffle(colors)
    random.seed(None)

    out_boxes, out_scores, out_classes, num_boxes = bboxes
    for i in range(num_boxes[0]):
        if int(out_classes[0][i]) < 0 or int(out_classes[0][i]) > num_classes:
            continue
        coor = out_boxes[0][i]
        coor[0] = int(coor[0] * image_h)
        coor[2] = int(coor[2] * image_h)
        coor[1] = int(coor[1] * image_w)
        coor[3] = int(coor[3] * image_w)

        fontScale = 0.5
        score = out_scores[0][i]
        class_ind = int(out_classes[0][i])
        bbox_color = colors[class_ind]
        bbox_thick = int(0.6 * (image_h + image_w) / 600)
        c1, c2 = (coor[1], coor[0]), (coor[3], coor[2])
        cv2.rectangle(image, c1, c2, bbox_color, bbox_thick)

        if show_label:
            bbox_mess = '%s: %.2f' % (classes[class_ind], score)
            t_size = cv2.getTextSize(
                bbox_mess, 0, fontScale, thickness=bbox_thick // 2)[0]
            c3 = (c1[0] + t_size[0], c1[1] - t_size[1] - 3)
            cv2.rectangle(image, c1, (np.float32(c3[0]), np.float32(
                c3[1])), bbox_color, -1)  # filled

            cv2.putText(image, bbox_mess, (c1[0], np.float32(c1[1] - 2)), cv2.FONT_HERSHEY_SIMPLEX,
                        fontScale, (0, 0, 0), bbox_thick // 2, lineType=cv2.LINE_AA)
    return image


def bbox_iou(bboxes1, bboxes2):
    """
    @param bboxes1: (a, b, ..., 4)
    @param bboxes2: (A, B, ..., 4)
        x:X is 1:n or n:n or n:1
    @return (max(a,A), max(b,B), ...)
    ex) (4,):(3,4) -> (3,)
        (2,1,4):(2,3,4) -> (2,3)
    """
    bboxes1_area = bboxes1[..., 2] * bboxes1[..., 3]
    bboxes2_area = bboxes2[..., 2] * bboxes2[..., 3]

    bboxes1_coor = tf.concat(
        [
            bboxes1[..., :2] - bboxes1[..., 2:] * 0.5,
            bboxes1[..., :2] + bboxes1[..., 2:] * 0.5,
        ],
        axis=-1,
    )
    bboxes2_coor = tf.concat(
        [
            bboxes2[..., :2] - bboxes2[..., 2:] * 0.5,
            bboxes2[..., :2] + bboxes2[..., 2:] * 0.5,
        ],
        axis=-1,
    )

    left_up = tf.maximum(bboxes1_coor[..., :2], bboxes2_coor[..., :2])
    right_down = tf.minimum(bboxes1_coor[..., 2:], bboxes2_coor[..., 2:])

    inter_section = tf.maximum(right_down - left_up, 0.0)
    inter_area = inter_section[..., 0] * inter_section[..., 1]

    union_area = bboxes1_area + bboxes2_area - inter_area

    iou = tf.math.divide_no_nan(inter_area, union_area)

    return iou


def bbox_giou(bboxes1, bboxes2):
    """
    Generalized IoU
    @param bboxes1: (a, b, ..., 4)
    @param bboxes2: (A, B, ..., 4)
        x:X is 1:n or n:n or n:1
    @return (max(a,A), max(b,B), ...)
    ex) (4,):(3,4) -> (3,)
        (2,1,4):(2,3,4) -> (2,3)
    """
    bboxes1_area = bboxes1[..., 2] * bboxes1[..., 3]
    bboxes2_area = bboxes2[..., 2] * bboxes2[..., 3]

    bboxes1_coor = tf.concat([bboxes1[..., :2] - bboxes1[..., 2:] * 0.5, bboxes1[..., :2] + bboxes1[..., 2:] * 0.5, ],
                             axis=-1, )
    bboxes2_coor = tf.concat([bboxes2[..., :2] - bboxes2[..., 2:] * 0.5, bboxes2[..., :2] + bboxes2[..., 2:] * 0.5, ],
                             axis=-1, )

    left_up = tf.maximum(bboxes1_coor[..., :2], bboxes2_coor[..., :2])
    right_down = tf.minimum(bboxes1_coor[..., 2:], bboxes2_coor[..., 2:])

    inter_section = tf.maximum(right_down - left_up, 0.0)
    inter_area = inter_section[..., 0] * inter_section[..., 1]

    union_area = bboxes1_area + bboxes2_area - inter_area

    iou = tf.math.divide_no_nan(inter_area, union_area)

    enclose_left_up = tf.minimum(bboxes1_coor[..., :2], bboxes2_coor[..., :2])
    enclose_right_down = tf.maximum(
        bboxes1_coor[..., 2:], bboxes2_coor[..., 2:])

    enclose_section = enclose_right_down - enclose_left_up
    enclose_area = enclose_section[..., 0] * enclose_section[..., 1]

    giou = iou - tf.math.divide_no_nan(enclose_area - union_area, enclose_area)

    return giou


def bbox_ciou(bboxes1, bboxes2):
    """
    Complete IoU
    @param bboxes1: (a, b, ..., 4)
    @param bboxes2: (A, B, ..., 4)
        x:X is 1:n or n:n or n:1
    @return (max(a,A), max(b,B), ...)
    ex) (4,):(3,4) -> (3,)
        (2,1,4):(2,3,4) -> (2,3)
    """
    bboxes1_area = bboxes1[..., 2] * bboxes1[..., 3]
    bboxes2_area = bboxes2[..., 2] * bboxes2[..., 3]

    bboxes1_coor = tf.concat(
        [
            bboxes1[..., :2] - bboxes1[..., 2:] * 0.5,
            bboxes1[..., :2] + bboxes1[..., 2:] * 0.5,
        ],
        axis=-1,
    )
    bboxes2_coor = tf.concat(
        [
            bboxes2[..., :2] - bboxes2[..., 2:] * 0.5,
            bboxes2[..., :2] + bboxes2[..., 2:] * 0.5,
        ],
        axis=-1,
    )

    left_up = tf.maximum(bboxes1_coor[..., :2], bboxes2_coor[..., :2])
    right_down = tf.minimum(bboxes1_coor[..., 2:], bboxes2_coor[..., 2:])

    inter_section = tf.maximum(right_down - left_up, 0.0)
    inter_area = inter_section[..., 0] * inter_section[..., 1]

    union_area = bboxes1_area + bboxes2_area - inter_area

    iou = tf.math.divide_no_nan(inter_area, union_area)

    enclose_left_up = tf.minimum(bboxes1_coor[..., :2], bboxes2_coor[..., :2])
    enclose_right_down = tf.maximum(
        bboxes1_coor[..., 2:], bboxes2_coor[..., 2:])

    enclose_section = enclose_right_down - enclose_left_up

    c_2 = enclose_section[..., 0] ** 2 + enclose_section[..., 1] ** 2

    center_diagonal = bboxes2[..., :2] - bboxes1[..., :2]

    rho_2 = center_diagonal[..., 0] ** 2 + center_diagonal[..., 1] ** 2

    diou = iou - tf.math.divide_no_nan(rho_2, c_2)

    v = ((tf.math.atan(tf.math.divide_no_nan(bboxes1[..., 2], bboxes1[..., 3]))
          - tf.math.atan(tf.math.divide_no_nan(bboxes2[..., 2], bboxes2[..., 3]))) * 2 / np.pi) ** 2

    alpha = tf.math.divide_no_nan(v, 1 - iou + v)

    ciou = diou - alpha * v

    return ciou


def nms(bboxes, iou_threshold, sigma=0.3, method='nms'):
    """
    :param bboxes: (xmin, ymin, xmax, ymax, score, class)

    Note: soft-nms, https://arxiv.org/pdf/1704.04503.pdf
          https://github.com/bharatsingh430/soft-nms
    """
    classes_in_img = list(set(bboxes[:, 5]))
    best_bboxes = []

    for cls in classes_in_img:
        cls_mask = (bboxes[:, 5] == cls)
        cls_bboxes = bboxes[cls_mask]

        while len(cls_bboxes) > 0:
            max_ind = np.argmax(cls_bboxes[:, 4])
            best_bbox = cls_bboxes[max_ind]
            best_bboxes.append(best_bbox)
            cls_bboxes = np.concatenate(
                [cls_bboxes[: max_ind], cls_bboxes[max_ind + 1:]])
            iou = bbox_iou(best_bbox[np.newaxis, :4], cls_bboxes[:, :4])
            weight = np.ones((len(iou),), dtype=np.float32)

            assert method in ['nms', 'soft-nms']

            if method == 'nms':
                iou_mask = iou > iou_threshold
                weight[iou_mask] = 0.0

            if method == 'soft-nms':
                weight = np.exp(-(1.0 * iou ** 2 / sigma))

            cls_bboxes[:, 4] = cls_bboxes[:, 4] * weight
            score_mask = cls_bboxes[:, 4] > 0.
            cls_bboxes = cls_bboxes[score_mask]

    return best_bboxes


def freeze_all(model, frozen=True):
    model.trainable = not frozen
    if isinstance(model, tf.keras.Model):
        for l in model.layers:
            freeze_all(l, frozen)


def unfreeze_all(model, frozen=False):
    model.trainable = not frozen
    if isinstance(model, tf.keras.Model):
        for l in model.layers:
            unfreeze_all(l, frozen)



import os


def read_lines(file_name):
    """loads class name from a file"""
    classes_path = os.path.expanduser(file_name)
    with open(classes_path, encoding='utf8') as f:
        class_names = f.readlines()
    class_names = [c.strip() for c in class_names]
    return class_names



from notekeras.backend import backend as K
from notekeras.backend import keras


class MaskedConv1D(keras.layers.Conv1D):

    def __init__(self, **kwargs):
        super(MaskedConv1D, self).__init__(**kwargs)
        self.supports_masking = True

    def compute_mask(self, inputs, mask=None):
        if mask is not None and self.padding == 'valid':
            mask = mask[:, self.kernel_size[0] // 2 * self.dilation_rate[0] * 2:]
        return mask

    def call(self, inputs, mask=None):
        if mask is not None:
            mask = K.cast(mask, K.floatx())
            inputs *= K.expand_dims(mask, axis=-1)
        return super(MaskedConv1D, self).call(inputs)



import datetime
import os
import pickle

import numpy as np
import tensorflow as tf
from notedata.dataset import ElectronicsData
from tensorflow import keras

from .model import DIN

electronic = ElectronicsData(data_path='./download/')


def download():
    electronic.download()
    electronic.preprocess()


def input_data(dataset, max_sl):
    user = np.array(dataset[:, 0], dtype='int32')
    item = np.array(dataset[:, 1], dtype='int32')
    hist = dataset[:, 2]
    hist_matrix = keras.preprocessing.sequence.pad_sequences(
        hist, maxlen=max_sl, padding='post')

    sl = np.array(dataset[:, 3], dtype='int32')
    y = np.array(dataset[:, 4], dtype='float32')

    return user, item, hist_matrix, sl, y


def train():
    hidden_unit = 64
    batch_size = 32
    learning_rate = 1
    epochs = 50

    with open(electronic.pkl_dataset, 'rb') as f:
        train_set = np.array(pickle.load(f))
        test_set = pickle.load(f)
        cate_list = pickle.load(f)
        user_count, item_count, cate_count, max_sl = pickle.load(f)
    train_user, train_item, train_hist, train_sl, train_y = input_data(
        train_set, max_sl)
    # Tensorboard
    current_time = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    log_dir = 'logs/' + current_time
    tensorboard = keras.callbacks.TensorBoard(
        log_dir=log_dir,
        histogram_freq=1,
        write_graph=True,
        write_grads=False,
        write_images=True,
        embeddings_freq=0, embeddings_layer_names=None,
        embeddings_metadata=None, embeddings_data=None, update_freq=500
    )
    # model checkpoint
    check_path = 'save/din_weights.epoch_{epoch:04d}.val_loss_{val_loss:.4f}.ckpt'
    checkpoint = tf.keras.callbacks.ModelCheckpoint(check_path, save_weights_only=True,
                                                    verbose=1, period=1)

    model = DIN(user_count, item_count, cate_count, cate_list, hidden_unit)
    model.summary()
    optimizer = keras.optimizers.SGD(learning_rate=learning_rate, decay=0.1)
    model.compile(loss=keras.losses.binary_crossentropy,
                  optimizer=optimizer, metrics=[tf.keras.metrics.AUC()])
    model.fit(
        [train_user, train_item, train_hist, train_sl],
        train_y,
        epochs=epochs,
        batch_size=batch_size,
        validation_split=0.1,
        callbacks=[tensorboard, checkpoint]
    )



from notekeras.layers import BatchNormalizationFreeze
from tensorflow import keras
from tensorflow.keras import layers

parameters = {
    "kernel_initializer": "he_normal"
}


def basic_1d(filters, stage=0, block=0, kernel_size=3, numerical_name=False, stride=None, freeze_bn=False):
    """
    A one-dimensional basic block.

    :param filters: the output’s feature space

    :param stage: int representing the stage of this block (starting from 0)

    :param block: int representing this block (starting from 0)

    :param kernel_size: size of the kernel

    :param numerical_name: if true, uses numbers to represent blocks instead of chars (ResNet{101, 152, 200})

    :param stride: int representing the stride used in the shortcut and the first conv layer, default derives stride from block id

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    Usage:

        >>> import keras_resnet.blocks

        >>> keras_resnet.blocks.basic_1d(64)
    """
    if stride is None:
        if block != 0 or stage == 0:
            stride = 1
        else:
            stride = 2

    if keras.backend.image_data_format() == "channels_last":
        axis = -1
    else:
        axis = 1

    if block > 0 and numerical_name:
        block_char = "b{}".format(block)
    else:
        block_char = chr(ord('a') + block)

    stage_char = str(stage + 2)

    def f(x):
        y = layers.ZeroPadding1D(
            padding=1,
            name="padding{}{}_branch2a".format(stage_char, block_char)
        )(x)

        y = layers.Conv1D(
            filters,
            kernel_size,
            strides=stride,
            use_bias=False,
            name="res{}{}_branch2a".format(stage_char, block_char),
            **parameters
        )(y)

        y = BatchNormalizationFreeze(
            axis=axis,
            epsilon=1e-5,
            freeze=freeze_bn,
            name="bn{}{}_branch2a".format(stage_char, block_char)
        )(y)

        y = layers.Activation(
            "relu",
            name="res{}{}_branch2a_relu".format(stage_char, block_char)
        )(y)

        y = layers.ZeroPadding1D(
            padding=1,
            name="padding{}{}_branch2b".format(stage_char, block_char)
        )(y)

        y = layers.Conv1D(
            filters,
            kernel_size,
            use_bias=False,
            name="res{}{}_branch2b".format(stage_char, block_char),
            **parameters
        )(y)

        y = BatchNormalizationFreeze(
            axis=axis,
            epsilon=1e-5,
            freeze=freeze_bn,
            name="bn{}{}_branch2b".format(stage_char, block_char)
        )(y)

        if block == 0:
            shortcut = layers.Conv1D(
                filters,
                1,
                strides=stride,
                use_bias=False,
                name="res{}{}_branch1".format(stage_char, block_char),
                **parameters
            )(x)

            shortcut = BatchNormalizationFreeze(
                axis=axis,
                epsilon=1e-5,
                freeze=freeze_bn,
                name="bn{}{}_branch1".format(stage_char, block_char)
            )(shortcut)
        else:
            shortcut = x

        y = layers.Add(
            name="res{}{}".format(stage_char, block_char)
        )([y, shortcut])

        y = layers.Activation(
            "relu",
            name="res{}{}_relu".format(stage_char, block_char)
        )(y)

        return y

    return f


def bottleneck_1d(filters, stage=0, block=0, kernel_size=3, numerical_name=False, stride=None, freeze_bn=False):
    """
    A one-dimensional bottleneck block.

    :param filters: the output’s feature space

    :param stage: int representing the stage of this block (starting from 0)

    :param block: int representing this block (starting from 0)

    :param kernel_size: size of the kernel

    :param numerical_name: if true, uses numbers to represent blocks instead of chars (ResNet{101, 152, 200})

    :param stride: int representing the stride used in the shortcut and the first conv layer, default derives stride from block id

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    Usage:

        >>> import keras_resnet.blocks

        >>> keras_resnet.blocks.bottleneck_1d(64)
    """
    if stride is None:
        stride = 1 if block != 0 or stage == 0 else 2

    if keras.backend.image_data_format() == "channels_last":
        axis = -1
    else:
        axis = 1

    if block > 0 and numerical_name:
        block_char = "b{}".format(block)
    else:
        block_char = chr(ord('a') + block)

    stage_char = str(stage + 2)

    def f(x):
        y = layers.Conv1D(
            filters,
            1,
            strides=stride,
            use_bias=False,
            name="res{}{}_branch2a".format(stage_char, block_char),
            **parameters
        )(x)

        y = BatchNormalizationFreeze(
            axis=axis,
            epsilon=1e-5,
            freeze=freeze_bn,
            name="bn{}{}_branch2a".format(stage_char, block_char)
        )(y)

        y = layers.Activation(
            "relu",
            name="res{}{}_branch2a_relu".format(stage_char, block_char)
        )(y)

        y = layers.ZeroPadding1D(
            padding=1,
            name="padding{}{}_branch2b".format(stage_char, block_char)
        )(y)

        y = layers.Conv1D(
            filters,
            kernel_size,
            use_bias=False,
            name="res{}{}_branch2b".format(stage_char, block_char),
            **parameters
        )(y)

        y = BatchNormalizationFreeze(
            axis=axis,
            epsilon=1e-5,
            freeze=freeze_bn,
            name="bn{}{}_branch2b".format(stage_char, block_char)
        )(y)

        y = layers.Activation(
            "relu", name="res{}{}_branch2b_relu".format(stage_char, block_char))(y)

        y = layers.Conv1D(filters * 4, 1, use_bias=False,
                          name="res{}{}_branch2c".format(stage_char, block_char), **parameters)(y)

        y = BatchNormalizationFreeze(
            axis=axis,
            epsilon=1e-5,
            freeze=freeze_bn,
            name="bn{}{}_branch2c".format(stage_char, block_char)
        )(y)

        if block == 0:
            shortcut = layers.Conv1D(
                filters * 4,
                1,
                strides=stride,
                use_bias=False,
                name="res{}{}_branch1".format(stage_char, block_char),
                **parameters
            )(x)

            shortcut = BatchNormalizationFreeze(
                axis=axis,
                epsilon=1e-5,
                freeze=freeze_bn,
                name="bn{}{}_branch1".format(stage_char, block_char)
            )(shortcut)
        else:
            shortcut = x

        y = layers.Add(
            name="res{}{}".format(stage_char, block_char)
        )([y, shortcut])

        y = layers.Activation(
            "relu",
            name="res{}{}_relu".format(stage_char, block_char)
        )(y)

        return y

    return f


def basic_2d(filters, stage=0, block=0, kernel_size=3, numerical_name=False, stride=None, freeze_bn=False):
    """
    A two-dimensional basic block.

    :param filters: the output’s feature space

    :param stage: int representing the stage of this block (starting from 0)

    :param block: int representing this block (starting from 0)

    :param kernel_size: size of the kernel

    :param numerical_name: if true, uses numbers to represent blocks instead of chars (ResNet{101, 152, 200})

    :param stride: int representing the stride used in the shortcut and the first conv layer, default derives stride from block id

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    Usage:

        >>> import keras_resnet.blocks

        >>> keras_resnet.blocks.basic_2d(64)
    """
    if stride is None:
        if block != 0 or stage == 0:
            stride = 1
        else:
            stride = 2

    if keras.backend.image_data_format() == "channels_last":
        axis = 3
    else:
        axis = 1

    if block > 0 and numerical_name:
        block_char = "b{}".format(block)
    else:
        block_char = chr(ord('a') + block)

    stage_char = str(stage + 2)

    def f(x):
        y = layers.ZeroPadding2D(
            padding=1, name="padding{}{}_branch2a".format(stage_char, block_char))(x)

        y = layers.Conv2D(filters, kernel_size, strides=stride, use_bias=False,
                          name="res{}{}_branch2a".format(stage_char, block_char), **parameters)(y)

        y = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                     name="bn{}{}_branch2a".format(stage_char, block_char))(y)

        y = layers.Activation(
            "relu", name="res{}{}_branch2a_relu".format(stage_char, block_char))(y)

        y = layers.ZeroPadding2D(
            padding=1, name="padding{}{}_branch2b".format(stage_char, block_char))(y)

        y = layers.Conv2D(filters, kernel_size, use_bias=False,
                          name="res{}{}_branch2b".format(stage_char, block_char), **parameters)(y)

        y = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                     name="bn{}{}_branch2b".format(stage_char, block_char))(y)

        if block == 0:
            shortcut = layers.Conv2D(filters, (1, 1), strides=stride, use_bias=False,
                                     name="res{}{}_branch1".format(stage_char, block_char), **parameters)(x)

            shortcut = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                                name="bn{}{}_branch1".format(stage_char, block_char))(
                shortcut)
        else:
            shortcut = x

        y = layers.Add(name="res{}{}".format(
            stage_char, block_char))([y, shortcut])

        y = layers.Activation(
            "relu", name="res{}{}_relu".format(stage_char, block_char))(y)

        return y

    return f


def bottleneck_2d(filters, stage=0, block=0, kernel_size=3, numerical_name=False, stride=None, freeze_bn=False):
    """
    A two-dimensional bottleneck block.

    :param filters: the output’s feature space

    :param stage: int representing the stage of this block (starting from 0)

    :param block: int representing this block (starting from 0)

    :param kernel_size: size of the kernel

    :param numerical_name: if true, uses numbers to represent blocks instead of chars (ResNet{101, 152, 200})

    :param stride: int representing the stride used in the shortcut and the first conv layer, default derives stride from block id

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    Usage:

        >>> import keras_resnet.blocks

        >>> keras_resnet.blocks.bottleneck_2d(64)
    """
    if stride is None:
        if block != 0 or stage == 0:
            stride = 1
        else:
            stride = 2

    if keras.backend.image_data_format() == "channels_last":
        axis = 3
    else:
        axis = 1

    if block > 0 and numerical_name:
        block_char = "b{}".format(block)
    else:
        block_char = chr(ord('a') + block)

    stage_char = str(stage + 2)

    def f(x):
        y = layers.Conv2D(filters, (1, 1), strides=stride, use_bias=False,
                          name="res{}{}_branch2a".format(stage_char, block_char), **parameters)(x)

        y = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                     name="bn{}{}_branch2a".format(stage_char, block_char))(y)

        y = layers.Activation(
            "relu", name="res{}{}_branch2a_relu".format(stage_char, block_char))(y)

        y = layers.ZeroPadding2D(
            padding=1, name="padding{}{}_branch2b".format(stage_char, block_char))(y)

        y = layers.Conv2D(filters, kernel_size, use_bias=False,
                          name="res{}{}_branch2b".format(stage_char, block_char), **parameters)(y)

        y = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                     name="bn{}{}_branch2b".format(stage_char, block_char))(y)

        y = layers.Activation(
            "relu", name="res{}{}_branch2b_relu".format(stage_char, block_char))(y)

        y = layers.Conv2D(filters * 4, (1, 1), use_bias=False,
                          name="res{}{}_branch2c".format(stage_char, block_char), **parameters)(y)

        y = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                     name="bn{}{}_branch2c".format(stage_char, block_char))(y)

        if block == 0:
            shortcut = layers.Conv2D(filters * 4, (1, 1), strides=stride, use_bias=False,
                                     name="res{}{}_branch1".format(stage_char, block_char), **parameters)(x)

            shortcut = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                                name="bn{}{}_branch1".format(stage_char, block_char))(
                shortcut)
        else:
            shortcut = x

        y = layers.Add(name="res{}{}".format(
            stage_char, block_char))([y, shortcut])

        y = layers.Activation(
            "relu", name="res{}{}_relu".format(stage_char, block_char))(y)

        return y

    return f


def basic_3d(filters, stage=0, block=0, kernel_size=3, numerical_name=False, stride=None, freeze_bn=False):
    """
    A three-dimensional basic block.

    :param filters: the output’s feature space

    :param stage: int representing the stage of this block (starting from 0)

    :param block: int representing this block (starting from 0)

    :param kernel_size: size of the kernel

    :param numerical_name: if true, uses numbers to represent blocks instead of chars (ResNet{101, 152, 200})

    :param stride: int representing the stride used in the shortcut and the first conv layer, default derives stride from block id

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    Usage:

        >>> import keras_resnet.blocks

        >>> keras_resnet.blocks.basic_3d(64)
    """
    if stride is None:
        if block != 0 or stage == 0:
            stride = 1
        else:
            stride = 2

    if keras.backend.image_data_format() == "channels_last":
        axis = 3
    else:
        axis = 1

    if block > 0 and numerical_name:
        block_char = "b{}".format(block)
    else:
        block_char = chr(ord('a') + block)

    stage_char = str(stage + 2)

    def f(x):
        y = layers.ZeroPadding3D(
            padding=1, name="padding{}{}_branch2a".format(stage_char, block_char))(x)

        y = layers.Conv3D(filters, kernel_size, strides=stride, use_bias=False,
                          name="res{}{}_branch2a".format(stage_char, block_char), **parameters)(y)

        y = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                     name="bn{}{}_branch2a".format(stage_char, block_char))(y)

        y = layers.Activation(
            "relu", name="res{}{}_branch2a_relu".format(stage_char, block_char))(y)

        y = layers.ZeroPadding3D(
            padding=1, name="padding{}{}_branch2b".format(stage_char, block_char))(y)

        y = layers.Conv3D(filters, kernel_size, use_bias=False,
                          name="res{}{}_branch2b".format(stage_char, block_char), **parameters)(y)

        y = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                     name="bn{}{}_branch2b".format(stage_char, block_char))(y)

        if block == 0:
            shortcut = layers.Conv3D(filters, (1, 1), strides=stride, use_bias=False,
                                     name="res{}{}_branch1".format(stage_char, block_char), **parameters)(x)

            shortcut = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                                name="bn{}{}_branch1".format(stage_char, block_char))(
                shortcut)
        else:
            shortcut = x

        y = layers.Add(name="res{}{}".format(
            stage_char, block_char))([y, shortcut])

        y = layers.Activation(
            "relu", name="res{}{}_relu".format(stage_char, block_char))(y)

        return y

    return f


def bottleneck_3d(filters, stage=0, block=0, kernel_size=3, numerical_name=False, stride=None, freeze_bn=False):
    """
    A three-dimensional bottleneck block.

    :param filters: the output’s feature space

    :param stage: int representing the stage of this block (starting from 0)

    :param block: int representing this block (starting from 0)

    :param kernel_size: size of the kernel

    :param numerical_name: if true, uses numbers to represent blocks instead of chars (ResNet{101, 152, 200})

    :param stride: int representing the stride used in the shortcut and the first conv layer, default derives stride from block id

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    Usage:

        >>> import keras_resnet.blocks

        >>> keras_resnet.blocks.bottleneck_3d(64)
    """
    if stride is None:
        if block != 0 or stage == 0:
            stride = 1
        else:
            stride = 2

    if keras.backend.image_data_format() == "channels_last":
        axis = 3
    else:
        axis = 1

    if block > 0 and numerical_name:
        block_char = "b{}".format(block)
    else:
        block_char = chr(ord('a') + block)

    stage_char = str(stage + 2)

    def f(x):
        y = layers.Conv3D(filters, (1, 1), strides=stride, use_bias=False,
                          name="res{}{}_branch2a".format(stage_char, block_char), **parameters)(x)

        y = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                     name="bn{}{}_branch2a".format(stage_char, block_char))(y)

        y = layers.Activation(
            "relu", name="res{}{}_branch2a_relu".format(stage_char, block_char))(y)

        y = layers.ZeroPadding3D(
            padding=1, name="padding{}{}_branch2b".format(stage_char, block_char))(y)

        y = layers.Conv3D(filters, kernel_size, use_bias=False,
                          name="res{}{}_branch2b".format(stage_char, block_char), **parameters)(y)

        y = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                     name="bn{}{}_branch2b".format(stage_char, block_char))(y)

        y = layers.Activation(
            "relu", name="res{}{}_branch2b_relu".format(stage_char, block_char))(y)

        y = layers.Conv3D(filters * 4, (1, 1), use_bias=False,
                          name="res{}{}_branch2c".format(stage_char, block_char), **parameters)(y)

        y = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                     name="bn{}{}_branch2c".format(stage_char, block_char))(y)

        if block == 0:
            shortcut = layers.Conv3D(filters * 4, (1, 1), strides=stride, use_bias=False,
                                     name="res{}{}_branch1".format(stage_char, block_char), **parameters)(x)

            shortcut = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn,
                                                name="bn{}{}_branch1".format(stage_char, block_char))(
                shortcut)
        else:
            shortcut = x

        y = layers.Add(name="res{}{}".format(
            stage_char, block_char))([y, shortcut])

        y = layers.Activation(
            "relu", name="res{}{}_relu".format(stage_char, block_char))(y)

        return y

    return f


def time_distributed_basic_2d(filters, stage=0, block=0, kernel_size=3, numerical_name=False, stride=None,
                              freeze_bn=False):
    """

    A time distributed two-dimensional basic block.

    :param filters: the output’s feature space

    :param stage: int representing the stage of this block (starting from 0)

    :param block: int representing this block (starting from 0)

    :param kernel_size: size of the kernel

    :param numerical_name: if true, uses numbers to represent blocks instead of chars (ResNet{101, 152, 200})

    :param stride: int representing the stride used in the shortcut and the first conv layer, default derives stride from block id

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    Usage:

        >>> import keras_resnet.blocks

        >>> keras_resnet.blocks.time_distributed_basic_2d(64)

    """
    if stride is None:
        if block != 0 or stage == 0:
            stride = 1
        else:
            stride = 2

    if keras.backend.image_data_format() == "channels_last":
        axis = 3
    else:
        axis = 1

    if block > 0 and numerical_name:
        block_char = "b{}".format(block)
    else:
        block_char = chr(ord('a') + block)

    stage_char = str(stage + 2)

    def f(x):
        y = layers.TimeDistributed(layers.ZeroPadding2D(padding=1),
                                   name="padding{}{}_branch2a".format(stage_char, block_char))(x)

        y = layers.TimeDistributed(
            layers.Conv2D(filters, kernel_size,
                          strides=stride, use_bias=False, **parameters),
            name="res{}{}_branch2a".format(stage_char, block_char))(y)

        y = layers.TimeDistributed(
            BatchNormalizationFreeze(
                axis=axis, epsilon=1e-5, freeze=freeze_bn),
            name="bn{}{}_branch2a".format(stage_char, block_char))(y)

        y = layers.TimeDistributed(layers.Activation("relu"),
                                   name="res{}{}_branch2a_relu".format(stage_char, block_char))(y)

        y = layers.TimeDistributed(layers.ZeroPadding2D(padding=1),
                                   name="padding{}{}_branch2b".format(stage_char, block_char))(y)

        y = layers.TimeDistributed(layers.Conv2D(filters, kernel_size, use_bias=False, **parameters),
                                   name="res{}{}_branch2b".format(stage_char, block_char))(y)

        y = layers.TimeDistributed(
            BatchNormalizationFreeze(
                axis=axis, epsilon=1e-5, freeze=freeze_bn),
            name="bn{}{}_branch2b".format(stage_char, block_char))(y)

        if block == 0:
            shortcut = layers.TimeDistributed(
                layers.Conv2D(
                    filters, (1, 1), strides=stride, use_bias=False, **parameters),
                name="res{}{}_branch1".format(stage_char, block_char))(x)

            shortcut = layers.TimeDistributed(
                BatchNormalizationFreeze(
                    axis=axis, epsilon=1e-5, freeze=freeze_bn),
                name="bn{}{}_branch1".format(stage_char, block_char))(shortcut)
        else:
            shortcut = x

        y = layers.Add(name="res{}{}".format(
            stage_char, block_char))([y, shortcut])

        y = layers.TimeDistributed(layers.Activation("relu"),
                                   name="res{}{}_relu".format(stage_char, block_char))(y)

        return y

    return f


def time_distributed_bottleneck_2d(filters, stage=0, block=0, kernel_size=3, numerical_name=False, stride=None,
                                   freeze_bn=False):
    """

    A time distributed two-dimensional bottleneck block.

    :param filters: the output’s feature space

    :param stage: int representing the stage of this block (starting from 0)

    :param block: int representing this block (starting from 0)

    :param kernel_size: size of the kernel

    :param numerical_name: if true, uses numbers to represent blocks instead of chars (ResNet{101, 152, 200})

    :param stride: int representing the stride used in the shortcut and the first conv layer, default derives stride from block id

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    Usage:

        >>> import keras_resnet.blocks

        >>> keras_resnet.blocks.time_distributed_bottleneck_2d(64)

    """
    if stride is None:
        if block != 0 or stage == 0:
            stride = 1
        else:
            stride = 2

    if keras.backend.image_data_format() == "channels_last":
        axis = 3
    else:
        axis = 1

    if block > 0 and numerical_name:
        block_char = "b{}".format(block)
    else:
        block_char = chr(ord('a') + block)

    stage_char = str(stage + 2)

    def f(x):
        y = layers.TimeDistributed(
            layers.Conv2D(filters, (1, 1), strides=stride,
                          use_bias=False, **parameters),
            name="res{}{}_branch2a".format(stage_char, block_char))(x)

        y = layers.TimeDistributed(
            BatchNormalizationFreeze(
                axis=axis, epsilon=1e-5, freeze=freeze_bn),
            name="bn{}{}_branch2a".format(stage_char, block_char))(y)

        y = layers.TimeDistributed(layers.Activation("relu"),
                                   name="res{}{}_branch2a_relu".format(stage_char, block_char))(y)

        y = layers.TimeDistributed(layers.ZeroPadding2D(padding=1),
                                   name="padding{}{}_branch2b".format(stage_char, block_char))(y)

        y = layers.TimeDistributed(layers.Conv2D(filters, kernel_size, use_bias=False, **parameters),
                                   name="res{}{}_branch2b".format(stage_char, block_char))(y)

        y = layers.TimeDistributed(
            BatchNormalizationFreeze(
                axis=axis, epsilon=1e-5, freeze=freeze_bn),
            name="bn{}{}_branch2b".format(stage_char, block_char))(y)

        y = layers.TimeDistributed(layers.Activation("relu"),
                                   name="res{}{}_branch2b_relu".format(stage_char, block_char))(y)

        y = layers.TimeDistributed(layers.Conv2D(filters * 4, (1, 1), use_bias=False, **parameters),
                                   name="res{}{}_branch2c".format(stage_char, block_char))(y)

        y = layers.TimeDistributed(
            BatchNormalizationFreeze(
                axis=axis, epsilon=1e-5, freeze=freeze_bn),
            name="bn{}{}_branch2c".format(stage_char, block_char))(y)

        if block == 0:
            shortcut = layers.TimeDistributed(
                layers.Conv2D(
                    filters * 4, (1, 1), strides=stride, use_bias=False, **parameters),
                name="res{}{}_branch1".format(stage_char, block_char))(x)

            shortcut = layers.TimeDistributed(
                layers.BatchNormalization(
                    axis=axis, epsilon=1e-5, freeze=freeze_bn),
                name="bn{}{}_branch1".format(stage_char, block_char))(shortcut)
        else:
            shortcut = x

        y = layers.Add(name="res{}{}".format(
            stage_char, block_char))([y, shortcut])

        y = layers.TimeDistributed(layers.Activation("relu"),
                                   name="res{}{}_relu".format(stage_char, block_char))(y)

        return y

    return f



import numpy as np
from notekeras.initializers import PriorProbability
from notekeras.layers.retinanet import (ClipBoxes, FilterDetections,
                                        RegressBoxes, UpSampleLike)
from notekeras.utils.image import read_image_batch
from tensorflow import keras
from tensorflow.keras import backend as K
from tensorflow.keras.layers import (Activation, Add, Concatenate, Conv2D,
                                     Input, Permute, Reshape)
from tensorflow.keras.models import Model

from .. import layers
from ..utils.anchors import AnchorParameters, anchor_parameter_default
from ..utils.image import preprocess_image, resize_image


class RetinaNetModel(Model):
    def __init__(self, inputs, backbone_layers, num_classes, num_anchors=None, submodels=None, name='retinanet'):
        """ Construct a RetinaNet model on top of a backbone.

        This model is the minimum model necessary for training (with the unfortunate exception of anchors as output).

        Args
            inputs                  : keras.layers.Input (or list of) for the input to the model.
            num_classes             : Number of classes to classify.
            num_anchors             : Number of base anchors.
            create_pyramid_features : Functor for creating pyramid features given the features C3, C4, C5 from the backbone.
            submodels               : Submodels to run on each feature map (default is regression and classification submodels).
            name                    : Name of the model.

        Returns
            A keras.models.Model which takes an image as input and outputs generated anchors and the result from each submodel on every pyramid level.

            The order of the outputs is as defined in submodels:
            ```
            [
                regression, classification, other[0], other[1], ...
            ]
            ```
        """

        super(RetinaNetModel, self).__init__()
        if num_anchors is None:
            num_anchors = AnchorParameters.default.num_anchors()

        self.num_classes = num_classes
        self.num_anchors = num_anchors
        if submodels is None:
            submodels = self.default_submodels()

        C3, C4, C5 = backbone_layers

        # compute pyramid features as per https://arxiv.org/abs/1708.02002
        features = self.create_pyramid_features(C3, C4, C5)

        # for all pyramid levels, run available submodels
        pyramids = self.build_pyramid(submodels, features)

        super(RetinaNetModel, self).__init__(inputs=inputs, outputs=pyramids, name=name)

    def build_pyramid(self, models, features):
        """ Applies all submodels to each FPN level.

        Args
            models   : List of submodels to run on each pyramid level (by default only regression, classifcation).
            features : The FPN features.

        Returns
            A list of tensors, one for each submodel.
        """
        return [self.build_model_pyramid(n, m, features) for n, m in models]

    def build_model_pyramid(self, name, model, features):
        """ Applies a single submodel to each FPN level.
        Args
            name     : Name of the submodel.
            model    : The submodel to evaluate.
            features : The FPN features.

        Returns
            A tensor containing the response from the submodel on the FPN features.
        """
        return Concatenate(axis=1, name=name)([model(f) for f in features])

    def create_pyramid_features(self, C3, C4, C5, feature_size=256):
        """ Creates the FPN layers on top of the backbone features.

        Args
            C3           : Feature stage C3 from the backbone.
            C4           : Feature stage C4 from the backbone.
            C5           : Feature stage C5 from the backbone.
            feature_size : The feature size to use for the resulting feature levels.

        Returns
            A list of feature levels [P3, P4, P5, P6, P7].
        """
        # upsample C5 to get P5 from the FPN paper
        P5 = Conv2D(feature_size, kernel_size=1, strides=1, padding='same', name='C5_reduced')(C5)
        P5_upsampled = UpSampleLike(name='P5_upsampled')([P5, C4])
        P5 = Conv2D(feature_size, kernel_size=3, strides=1, padding='same', name='P5')(P5)

        # add P5 elementwise to C4
        P4 = Conv2D(feature_size, kernel_size=1, strides=1, padding='same', name='C4_reduced')(C4)
        P4 = Add(name='P4_merged')([P5_upsampled, P4])
        P4_upsampled = UpSampleLike(name='P4_upsampled')([P4, C3])
        P4 = Conv2D(feature_size, kernel_size=3, strides=1, padding='same', name='P4')(P4)

        # add P4 elementwise to C3
        P3 = Conv2D(feature_size, kernel_size=1, strides=1, padding='same', name='C3_reduced')(C3)
        P3 = Add(name='P3_merged')([P4_upsampled, P3])
        P3 = Conv2D(feature_size, kernel_size=3, strides=1, padding='same', name='P3')(P3)

        # "P6 is obtained via a 3x3 stride-2 conv on C5"
        P6 = Conv2D(feature_size, kernel_size=3, strides=2, padding='same', name='P6')(C5)

        # "P7 is computed by applying ReLU followed by a 3x3 stride-2 conv on P6"
        P7 = Activation('relu', name='C6_relu')(P6)
        P7 = Conv2D(feature_size, kernel_size=3, strides=2, padding='same', name='P7')(P7)

        return [P3, P4, P5, P6, P7]

    def default_submodels(self):
        """ Create a list of default submodels used for object detection.

        The default submodels contains a regression submodel and a classification submodel.

        Args
            num_classes : Number of classes to use.
            num_anchors : Number of base anchors.

        Returns
            A list of tuple, where the first element is the name of the submodel and the second element is the submodel itself.
        """
        return [('regression', self.default_regression_model(4)),
                ('classification', self.default_classification_model())]

    def default_regression_model(self, num_values, pyramid_feature_size=256, regression_feature_size=256,
                                 name='regression_submodel'):
        """ Creates the default regression submodel.

        Args
            num_values              : Number of values to regress.
            num_anchors             : Number of anchors to regress for each feature level.
            pyramid_feature_size    : The number of filters to expect from the feature pyramid levels.
            regression_feature_size : The number of filters to use in the layers in the regression submodel.
            name                    : The name of the submodel.

        Returns
            A keras.models.Model that predicts regression values for each anchor.
        """
        # All new conv layers except the final one in the
        # RetinaNet (classification) subnets are initialized
        # with bias b = 0 and a Gaussian weight fill with stddev = 0.01.
        options = {
            'kernel_size': 3,
            'strides': 1,
            'padding': 'same',
            'kernel_initializer': keras.initializers.glorot_normal(seed=None),
            'bias_initializer': 'zeros'
        }

        if K.image_data_format() == 'channels_first':
            inputs = Input(shape=(pyramid_feature_size, None, None))
        else:
            inputs = Input(shape=(None, None, pyramid_feature_size))
        outputs = inputs
        for i in range(4):
            outputs = Conv2D(
                filters=regression_feature_size,
                activation='relu',
                name='pyramid_regression_{}'.format(i),
                **options
            )(outputs)

        outputs = Conv2D(self.num_anchors * num_values, name='pyramid_regression', **options)(outputs)
        if K.image_data_format() == 'channels_first':
            outputs = Permute((2, 3, 1), name='pyramid_regression_permute')(outputs)
        outputs = Reshape((-1, num_values), name='pyramid_regression_reshape')(outputs)

        return Model(inputs=inputs, outputs=outputs, name=name)

    def default_classification_model(self, pyramid_feature_size=256, prior_probability=0.01,
                                     classification_feature_size=256, name='classification_submodel'):
        """ Creates the default classification submodel.

        Args
            num_classes                 : Number of classes to predict a score for at each feature level.
            num_anchors                 : Number of anchors to predict classification scores for at each feature level.
            pyramid_feature_size        : The number of filters to expect from the feature pyramid levels.
            classification_feature_size : The number of filters to use in the layers in the classification submodel.
            name                        : The name of the submodel.

        Returns
            A keras.models.Model that predicts classes for each anchor.
        """
        options = {'kernel_size': 3, 'strides': 1, 'padding': 'same', }

        if K.image_data_format() == 'channels_first':
            inputs = Input(shape=(pyramid_feature_size, None, None))
        else:
            inputs = Input(shape=(None, None, pyramid_feature_size))
        outputs = inputs
        for i in range(4):
            outputs = Conv2D(filters=classification_feature_size,
                             activation='relu',
                             name='pyramid_classification_{}'.format(i),
                             kernel_initializer=keras.initializers.glorot_normal(seed=None),
                             bias_initializer='zeros',
                             **options
                             )(outputs)

        outputs = Conv2D(filters=self.num_classes * self.num_anchors,
                         kernel_initializer=keras.initializers.glorot_normal(seed=None),
                         bias_initializer=PriorProbability(probability=prior_probability),
                         name='pyramid_classification',
                         **options
                         )(outputs)

        # reshape output and apply sigmoid
        if K.image_data_format() == 'channels_first':
            outputs = Permute((2, 3, 1), name='pyramid_classification_permute')(outputs)
        outputs = Reshape((-1, self.num_classes), name='pyramid_classification_reshape')(outputs)
        outputs = Activation('sigmoid', name='pyramid_classification_sigmoid')(outputs)

        return Model(inputs=inputs, outputs=outputs, name=name)


class RetinaNetBox(Model):
    def __init__(self, model=None, nms=True, class_specific_filter=True, name='retinanet-bbox', anchor_params=None,
                 nms_threshold=0.5, score_threshold=0.05, max_detections=300, parallel_iterations=32, **kwargs):
        """ Construct a RetinaNet model on top of a backbone and adds convenience functions to output boxes directly.

        This model uses the minimum retinanet model and appends a few layers to compute boxes within the graph.
        These layers include applying the regression values to the anchors and performing NMS.

        Args
            model                 : RetinaNet model to append bbox layers to. If None, it will create a RetinaNet model using **kwargs.
            nms                   : Whether to use non-maximum suppression for the filtering step.
            class_specific_filter : Whether to use class specific filtering or filter for the best scoring class only.
            name                  : Name of the model.
            anchor_params         : Struct containing anchor parameters. If None, default values are used.
            nms_threshold         : Threshold for the IoU value to determine when a box should be suppressed.
            score_threshold       : Threshold used to prefilter the boxes with.
            max_detections        : Maximum number of detections to keep.
            parallel_iterations   : Number of batch items to process in parallel.
            **kwargs              : Additional kwargs to pass to the minimal retinanet model.

        Returns
            A keras.models.Model which takes an image as input and outputs the detections on the image.

            The order is defined as follows:
            ```
            [
                boxes, scores, labels, other[0], other[1], ...
            ]
            ```
        """
        super(RetinaNetBox, self).__init__(name=name)

        anchor_params = anchor_params or anchor_parameter_default

        # create RetinaNet model
        if model is None:
            model = RetinaNetModel(num_anchors=anchor_params.num_anchors(), **kwargs)

        # compute the anchors
        features = [model.get_layer(p_name).output for p_name in ['P3', 'P4', 'P5', 'P6', 'P7']]
        anchors = self.build_anchors(anchor_params, features)

        # we expect the anchors, regression and classification values as first output
        regression = model.outputs[0]
        classification = model.outputs[1]

        # "other" can be any additional output from custom submodels, by default this will be []
        other = model.outputs[2:]

        # apply predicted regression to anchors
        boxes = RegressBoxes(name='boxes')([anchors, regression])
        boxes = ClipBoxes(name='clipped_boxes')([model.inputs[0], boxes])

        # filter detections (apply NMS / score threshold / select top-k)
        detections = FilterDetections(nms=nms,
                                      class_specific_filter=class_specific_filter,
                                      name='filtered_detections',
                                      nms_threshold=nms_threshold,
                                      score_threshold=score_threshold,
                                      max_detections=max_detections,
                                      parallel_iterations=parallel_iterations
                                      )([boxes, classification] + other)

        # construct the model
        super(RetinaNetBox, self).__init__(inputs=model.inputs, outputs=detections, name=name)

    def build_anchors(self, anchor_parameters, features):
        """ Builds anchors for the shape of the features from FPN.

        Args
            anchor_parameters : Parameteres that determine how anchors are generated.
            features          : The FPN features.

        Returns
            A tensor containing the anchors for the FPN features.

            The shape is:
            ```
            (batch_size, num_anchors, 4)
            ```
        """
        anchors = [
            layers.Anchors(size=anchor_parameters.sizes[i],
                           stride=anchor_parameters.strides[i],
                           ratios=anchor_parameters.ratios,
                           scales=anchor_parameters.scales,
                           name='anchors_{}'.format(i)
                           )(f) for i, f in enumerate(features)
        ]

        return Concatenate(axis=1, name='anchors')(anchors)

    def predict_result_batch(self, image_paths):
        images = read_image_batch(image_paths)

        scales = []
        image_inputs = []
        for image in images:
            # image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            image = preprocess_image(image)
            image, scale = resize_image(image)

            scales.append(scale)
            image_inputs.append(image)

        boxes, scores, labels = self.predict(np.array(image_inputs))

        for i in range(0, len(boxes)):
            boxes[i] = boxes[i] / scales[i]

        return boxes, scores, labels



from tensorflow import keras
from tensorflow.keras.models import Model

from notekeras.models.resnet import core as models


class ResNet18Classifier(Model):
    """
    A :class:`ResNet18 <ResNet18>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:
        >>> from notekeras.models.resnet.classifiers import ResNet18Classifier
        >>> shape, classes = (224, 224, 3), 1000
        >>> x = keras.layers.Input(shape)
        >>> model = ResNet18Classifier(x)
        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, classes):
        outputs = models.ResNet2D18(inputs)
        outputs = keras.layers.Flatten()(outputs.output)
        outputs = keras.layers.Dense(classes, activation="softmax")(outputs)
        super(ResNet18Classifier, self).__init__(inputs, outputs)


class ResNet34Classifier(Model):
    """
    A :class:`ResNet34 <ResNet34>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:
        >>> from notekeras.models.resnet.classifiers import ResNet34Classifier
        >>> shape, classes = (224, 224, 3), 1000
        >>> x = keras.layers.Input(shape)
        >>> model = ResNet34Classifier(x)
        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, classes):
        outputs = models.ResNet2D34(inputs)

        outputs = keras.layers.Flatten()(outputs.output)

        outputs = keras.layers.Dense(classes, activation="softmax")(outputs)

        super(ResNet34Classifier, self).__init__(inputs, outputs)


class ResNet50Classifier(Model):
    """
    A :class:`ResNet50 <ResNet50>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:

        >>> from notekeras.models.resnet.classifiers import ResNet50Classifier
        >>> shape, classes = (224, 224, 3), 1000
        >>> x = keras.layers.Input(shape)
        >>> model = ResNet50Classifier(x)
        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, classes):
        outputs = models.ResNet2D50(inputs)
        outputs = keras.layers.Flatten()(outputs.output)
        outputs = keras.layers.Dense(classes, activation="softmax")(outputs)
        super(ResNet50Classifier, self).__init__(inputs, outputs)


class ResNet101Classifier(Model):
    """
    A :class:`ResNet101 <ResNet101>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:
        >>> from notekeras.models.resnet.classifiers import ResNet101Classifier
        >>> shape, classes = (224, 224, 3), 1000
        >>> x = keras.layers.Input(shape)
        >>> model = ResNet101Classifier(x)
        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, classes):
        outputs = models.ResNet2D101(inputs)
        outputs = keras.layers.Flatten()(outputs.output)
        outputs = keras.layers.Dense(classes, activation="softmax")(outputs)
        super(ResNet101Classifier, self).__init__(inputs, outputs)


class ResNet152Classifier(Model):
    """
    A :class:`ResNet152 <ResNet152>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:
        >>> from notekeras.models.resnet.classifiers import ResNet152Classifier
        >>> shape, classes = (224, 224, 3), 1000
        >>> x = keras.layers.Input(shape)
        >>> model = ResNet152Classifier(x)
        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])

    """

    def __init__(self, inputs, classes):
        outputs = models.ResNet2D152(inputs)
        outputs = keras.layers.Flatten()(outputs.output)
        outputs = keras.layers.Dense(classes, activation="softmax")(outputs)
        super(ResNet152Classifier, self).__init__(inputs, outputs)


class ResNet200Classifier(Model):
    """
    A :class:`ResNet200 <ResNet200>` object.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    Usage:
        >>> from notekeras.models.resnet.classifiers import ResNet200Classifier
        >>> shape, classes = (224, 224, 3), 1000
        >>> x = keras.layers.Input(shape)
        >>> model = ResNet200Classifier(x)
        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, classes):
        outputs = models.ResNet2D200(inputs)
        outputs = keras.layers.Flatten()(outputs.output)
        outputs = keras.layers.Dense(classes, activation="softmax")(outputs)
        super(ResNet200Classifier, self).__init__(inputs, outputs)



import datetime
import pickle

from notedata.dataset.datas import CriteoData, ElectronicsData
from tensorflow.keras import (callbacks, losses, metrics, optimizers,
                              preprocessing)
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras.losses import binary_crossentropy
from tensorflow.keras.metrics import AUC
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.utils import plot_model

from notekeras.models.fm.model import *

from .model import *

criteo = CriteoData()
electronics = ElectronicsData()


def download(mode=1):
    criteo.download(mode=mode)
    electronics.download()


def train_all(mode=1, name='FM'):
    learning_rate = 0.001
    batch_size = 4096
    epochs = 10

    feature_columns, train, test = criteo.build_dataset(mode=mode)

    # ============================Build Model==========================
    model = None
    if name == 'FM':
        model = FM(feature_columns=feature_columns, k=10)
    elif name == 'AFM':
        mode = 'avg'  # 'max', 'avg'
        model = AFM(feature_columns, mode)
    elif name == 'FFM':
        model = FFM(feature_columns=feature_columns, k=10)
    elif name == "NFM":
        model = NFM(feature_columns, hidden_units=[
                    256, 128, 64], dnn_dropout=0.5)
    elif name == "DeepFM":
        model = DeepFM(feature_columns, k=10, hidden_units=[
                       256, 128, 64], dnn_dropout=0.5)
    elif name == "xDeepFM":
        model = xDeepFM(feature_columns, hidden_units=[
                        256, 128, 64], cin_size=[128, 128])
    elif name == "MF":
        model = MF(feature_columns, implicit=False, use_bias=True)
    elif name == "WideDeep":
        model = WideDeep(feature_columns, hidden_units=[
                         256, 128, 64], dnn_dropout=0.5)
    elif name == "DCN":
        model = DCN(feature_columns, [256, 128, 64], dnn_dropout=0.5)
    elif name == "PNN":
        model = PNN(feature_columns, hidden_units=[
                    256, 128, 64], dnn_dropout=0.5)
    elif name == "Deep_Crossing":
        model = Deep_Crossing(feature_columns, hidden_units=[256, 128, 64])

    if model is not None:
        model.compile(loss=binary_crossentropy, optimizer=Adam(
            learning_rate=learning_rate), metrics=[AUC()])

    # model.summary()
    # ==============================Fit==============================

    if model is not None:
        plot_model(model, to_file=criteo.path_root +
                   '/'+name+'.png', show_shapes=True)
        plot_model(model, to_file=criteo.path_root +
                   '/'+name+'-extend.png',
                   show_shapes=True, expand_nested=True)

        check_path = criteo.path_root+'/save/{name}_weights.epoch_{epoch}.val_loss.ckpt'.format(
            epoch=epochs, name=name)
        checkpoint = callbacks.ModelCheckpoint(
            check_path, save_weights_only=True, verbose=1, period=5)
        early = EarlyStopping(monitor='val_loss',
                              patience=2, restore_best_weights=True)
        train_x, train_y = train
        test_x, test_y = test
        model.fit(train_x, train_y, epochs=epochs,
                  batch_size=batch_size,
                  validation_split=0.1,
                  callbacks=[
                      # checkpoint,
                      early],
                  )

        print('test AUC: %f' % model.evaluate(test_x, test_y)[1])


def input_data(dataset, max_sl):
    user = np.array(dataset[:, 0], dtype='int32')
    item = np.array(dataset[:, 1], dtype='int32')
    hist = dataset[:, 2]
    hist_matrix = preprocessing.sequence.pad_sequences(
        hist, maxlen=max_sl, padding='post')

    sl = np.array(dataset[:, 3], dtype='int32')
    y = np.array(dataset[:, 4], dtype='float32')

    return user, item, hist_matrix, sl, y


def train_din(mode=1):
    hidden_unit = 64
    batch_size = 32
    learning_rate = 1
    epochs = 50

    with open(electronics.pkl_dataset, 'rb') as f:
        train_set = np.array(pickle.load(f))
        test_set = pickle.load(f)
        cate_list = pickle.load(f)
        user_count, item_count, cate_count, max_sl = pickle.load(f)
    train_user, train_item, train_hist, train_sl, train_y = input_data(
        train_set, max_sl)
    test_user, test_item, test_hist, test_sl, test_y = input_data(
        test_set, max_sl)

    current_time = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    log_dir = 'logs/' + current_time
    tensorboard = callbacks.TensorBoard(log_dir=log_dir,
                                        histogram_freq=1,
                                        write_graph=True,
                                        write_grads=False,
                                        write_images=True,
                                        embeddings_freq=0, embeddings_layer_names=None,
                                        embeddings_metadata=None, embeddings_data=None, update_freq=500
                                        )
    # model checkpoint
    check_path = 'save/din_weights.epoch_{epoch:04d}.val_loss_{val_loss:.4f}.ckpt'
    checkpoint = callbacks.ModelCheckpoint(
        check_path, save_weights_only=True, verbose=1, period=1)

    model = DIN(user_count, item_count, cate_count, cate_list, hidden_unit)
    model.summary()
    optimizer = optimizers.SGD(learning_rate=learning_rate, decay=0.1)
    model.compile(loss=losses.binary_crossentropy,
                  optimizer=optimizer, metrics=[metrics.AUC()])
    model.fit([train_user, train_item, train_hist, train_sl],
              train_y,
              epochs=epochs,
              batch_size=batch_size,
              validation_split=0.1,
              # callbacks=[tensorboard, checkpoint]
              )
    print('test AUC: %f' % model.evaluate(
        [test_user, test_item, test_hist, test_sl], test_y))



import numpy as np
from tensorflow.keras.layers import Dense, Layer
from tensorflow.keras.models import Model


class ComponentLayer(Layer):
    def __init__(self, inputs=None, layer_depth=1, *args, **kwargs):
        """
        :param layer_depth:
        :param as_model:组件是否当做一个模型返回，当模型的时候，输入inputs必须非空
        :param inputs:模型输入，当as_model为True时传入
        :param args:
        :param kwargs:
        """
        super(ComponentLayer, self).__init__(*args, **kwargs)

        self.layer_depth = layer_depth
        self.input_layer = inputs

        super(ComponentLayer, self).__init__(*args, **kwargs)

    def __call__(self, inputs, **kwargs):
        if self.layer_depth <= 0:
            return super(ComponentLayer, self).__call__(inputs=inputs, **kwargs)
        else:
            self.build(np.shape(inputs)[1:])
            return self.call(inputs, **kwargs)


class Component(Model):
    def __init__(self, as_model=False, inputs=None, layer_depth=1, *args, **kwargs):
        """
        :param layer_depth:
        :param as_model:组件是否当做一个模型返回，当模型的时候，输入inputs必须非空
        :param inputs:模型输入，当as_model为True时传入
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.as_model = as_model

        self.layer_depth = layer_depth
        self.input_layer = inputs

        if self.as_model:
            if inputs is None:
                raise BaseException("when as_model is True, inputs must be not None")
            self.build(np.shape(inputs)[1:])
            self.outputs = self.call(inputs)
            self.output_layer = self.outputs
            super().__init__(inputs=inputs, outputs=self.outputs, *args, **kwargs)
        else:
            super().__init__(*args, **kwargs)

    def __call__(self, inputs, **kwargs):
        if self.as_model:
            return super().__call__(inputs, **kwargs)

        if self.layer_depth <= 0:
            return super().__call__(inputs=inputs, **kwargs)
        else:
            self.build(np.shape(inputs)[1:])
            return self.call(inputs, **kwargs)


class AutoEncoder(Component):
    def __init__(self, name='AutoEncoder', encode_size=None, *args, **kwargs):
        self.encode_size = encode_size or [128, 64, 32]
        super(AutoEncoder, self).__init__(name=name, *args, **kwargs)

    def _build(self, inputs):
        output = inputs
        for size in self.encode_size:
            output = Dense(size, activation='tanh')(output)
        return output




from tensorflow.keras.layers import BatchNormalization
from tensorflow.keras.layers import Conv2D, Add, ZeroPadding2D, UpSampling2D, Concatenate
from tensorflow.keras.layers import LeakyReLU
from tensorflow.keras.regularizers import l2

from notekeras.component import Component
from notekeras.utils import compose

__all__ = ['DarknetConv2D', 'DarkConv', 'DarkConvSet', 'ResBlockBody', 'YoloModel']


class DarknetConv2D(Conv2D):
    def __init__(self, *args, **kwargs):
        darknet_conv_kwargs = {'kernel_regularizer': l2(5e-4),
                               'padding': 'valid' if kwargs.get('strides') == (2, 2) else 'same'}
        darknet_conv_kwargs.update(kwargs)

        super(DarknetConv2D, self).__init__(*args, **darknet_conv_kwargs)


class DarkConv(Component):
    def __init__(self, count, size, strides=(1, 1), *args, **kwargs):
        self.count = count
        self.size = size
        self.strides = strides
        self.compose_list = None
        super(DarkConv, self).__init__(*args, **kwargs)

    def build(self, input_shape):
        self.compose_list = [
            DarknetConv2D(self.count, self.size, use_bias=False, name='{}-conv'.format(self.name),
                          strides=self.strides),
            BatchNormalization(name='{}-norm'.format(self.name)),
            LeakyReLU(alpha=0.1, name='{}-relu'.format(self.name))
        ]

    def call(self, inputs, **kwargs):
        output = compose(*self.compose_list)(inputs)
        return output

    def compute_output_shape(self, input_shape):
        return self.conv.compute_output_shape(input_shape)


class DarkConvSet(Component):
    def __init__(self, num_filters, *args, **kwargs):
        self.num_filters = num_filters
        self.conv_list = None
        super(DarkConvSet, self).__init__(*args, **kwargs)

    def build(self, input_shape):
        self.conv_list = [
            DarkConv(self.num_filters, (1, 1), name='{}_1'.format(self.name)),
            DarkConv(self.num_filters * 2, (3, 3), name='{}_2'.format(self.name)),
            DarkConv(self.num_filters, (1, 1), name='{}_3'.format(self.name)),
            DarkConv(self.num_filters * 2, (3, 3), name='{}_4'.format(self.name)),
            DarkConv(self.num_filters, (1, 1), name='{}_5'.format(self.name))
        ]

    def call(self, inputs, **kwargs):
        return compose(*self.conv_list)(inputs)


class ResBlockBody(Component):
    def __init__(self, num_filters, num_blocks, name='ResBlockBody', *args, **kwargs):
        super(ResBlockBody, self).__init__(name=name)
        self.num_filters = num_filters
        self.num_blocks = num_blocks

        self.zero = None
        self.conv = None
        self.block_layers = None
        super(ResBlockBody, self).__init__(name=name, *args, **kwargs)

    def build(self, input_shape):
        self.zero = ZeroPadding2D(((1, 0), (1, 0)))
        self.conv = DarkConv(self.num_filters, (3, 3), strides=(2, 2), layer_depth=self.layer_depth - 1)
        self.block_layers = []
        for i in range(self.num_blocks):
            self.block_layers.append([
                DarkConv(self.num_filters // 2, (1, 1), layer_depth=self.layer_depth - 1),
                DarkConv(self.num_filters, (3, 3), layer_depth=self.layer_depth - 1),
                Add()
            ])

    def call(self, inputs, **kwargs):
        x = compose(self.zero, self.conv)(inputs)
        for conv1, conv2, add in self.block_layers:
            y = compose(conv1, conv2)(x)
            x = add([x, y])
        return x


class YoloModel(Component):
    def __init__(self, num_anchors, num_classes, name='Yolo', anchors=None, *args, **kwargs):
        self.num_anchors = num_anchors
        self.num_classes = num_classes
        self.anchors = anchors

        self.layer0 = self.block1 = self.block2 = self.block3 = self.block4 = self.block5 = None
        self.make_list = self.conv_samp = self.concat = None
        super(YoloModel, self).__init__(name=name, *args, **kwargs)

    def build(self, input_shape):
        self.layer0 = DarkConv(32, (3, 3), name='{}_DarkConv'.format(self.name), layer_depth=self.layer_depth - 2)

        self.block1 = ResBlockBody(64, 1, name='{}_ResBlock_1'.format(self.name), layer_depth=self.layer_depth - 1)
        self.block2 = ResBlockBody(128, 2, name='{}_ResBlock_2'.format(self.name), layer_depth=self.layer_depth - 1)
        self.block3 = ResBlockBody(256, 8, name='{}_ResBlock_3'.format(self.name), layer_depth=self.layer_depth - 1)
        self.block4 = ResBlockBody(512, 8, name='{}_ResBlock_4'.format(self.name), layer_depth=self.layer_depth - 1)
        self.block5 = ResBlockBody(1024, 4, name='{}_ResBlock_5'.format(self.name), layer_depth=self.layer_depth - 1)

        out_filters = self.num_anchors * (self.num_classes + 5)
        self.make_list = [
            [DarkConvSet(num_filters, layer_depth=self.layer_depth - 1),
             DarkConv(num_filters * 2, (3, 3), layer_depth=self.layer_depth - 1),
             DarknetConv2D(out_filters, (1, 1), name='yolo_conv2d_{}'.format(num_filters))
             ] for num_filters in [512, 256, 128]
        ]
        self.conv_samp = [[
            DarkConv(256, (1, 1), layer_depth=self.layer_depth - 2, name='{}__DarkConv_1'.format(self.name), ),
            UpSampling2D(2)
        ], [
            DarkConv(128, (1, 1), layer_depth=self.layer_depth - 2, name='{}__DarkConv_2'.format(self.name), ),
            UpSampling2D(2)
        ]
        ]
        self.concat = [Concatenate(), Concatenate()]

    def call(self, inputs, **kwargs):
        out0 = self.layer0(inputs)
        out1 = self.block1(out0)
        out2 = self.block2(out1)
        out3 = self.block3(out2)
        out4 = self.block4(out3)
        out5 = self.block5(out4)

        x = self.make_list[0][0](out5)
        y1 = compose(*self.make_list[0][1:])(x)

        x = compose(*self.conv_samp[0])(x)
        x = self.concat[0]([x, out4])

        x = self.make_list[1][0](x)
        y2 = compose(*self.make_list[1][1:])(x)

        x = compose(*self.conv_samp[1])(x)
        x = self.concat[1]([x, out3])

        x = self.make_list[2][0](x)
        y3 = compose(*self.make_list[2][1:])(x)

        return y1, y2, y3



import numpy as np
import tensorflow.keras.backend as K
from tensorflow.keras.layers import Layer

from notekeras.models.retinanet.utils import anchors as utils_anchors
from notekeras.utils import utils_retinanet


class Anchors(Layer):
    """ Keras layer for generating achors for a given shape.
    """

    def __init__(self, size, stride, ratios=None, scales=None, *args, **kwargs):
        """ Initializer for an Anchors layer.

        Args
            size: The base size of the anchors to generate.
            stride: The stride of the anchors to generate.
            ratios: The ratios of the anchors to generate (defaults to AnchorParameters.default.ratios).
            scales: The scales of the anchors to generate (defaults to AnchorParameters.default.scales).
        """
        self.size = size
        self.stride = stride
        self.ratios = ratios
        self.scales = scales

        if ratios is None:
            self.ratios = utils_anchors.AnchorParameters.default.ratios
        elif isinstance(ratios, list):
            self.ratios = np.array(ratios)
        if scales is None:
            self.scales = utils_anchors.AnchorParameters.default.scales
        elif isinstance(scales, list):
            self.scales = np.array(scales)

        self.num_anchors = len(self.ratios) * len(self.scales)
        self.anchors = K.variable(utils_anchors.generate_anchors(
            base_size=self.size,
            ratios=self.ratios,
            scales=self.scales,
        ))

        super(Anchors, self).__init__(*args, **kwargs)

    def call(self, inputs, **kwargs):
        features = inputs
        features_shape = K.shape(features)

        # generate proposals from bbox deltas and shifted anchors
        if K.image_data_format() == 'channels_first':
            anchors = utils_retinanet.shift(features_shape[2:4], self.stride, self.anchors)
        else:
            anchors = utils_retinanet.shift(features_shape[1:3], self.stride, self.anchors)
        anchors = K.tile(K.expand_dims(anchors, axis=0), (features_shape[0], 1, 1))

        return anchors

    def compute_output_shape(self, input_shape):
        if None not in input_shape[1:]:
            if K.image_data_format() == 'channels_first':
                total = np.prod(input_shape[2:4]) * self.num_anchors
            else:
                total = np.prod(input_shape[1:3]) * self.num_anchors

            return (input_shape[0], total, 4)
        else:
            return (input_shape[0], None, 4)

    def get_config(self):
        config = super(Anchors, self).get_config()
        config.update({
            'size': self.size,
            'stride': self.stride,
            'ratios': self.ratios.tolist(),
            'scales': self.scales.tolist(),
        })

        return config



from notekeras.component import resnet_bolck as B
from notekeras.layers import BatchNormalizationFreeze
from tensorflow import keras
from tensorflow.keras import layers
from tensorflow.keras.layers import Activation, Conv2D, MaxPooling2D
from tensorflow.keras.models import Model


class ResNet1D(Model):
    """
    Constructs a `keras.models.Model` object using the given block count.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param block: a residual block (e.g. an instance of `keras_resnet.blocks.basic_1d`)

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :param numerical_names: list of bool, same size as blocks, used to indicate whether names of layers should include numbers or letters

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.blocks
        >>> import keras_resnet.models
        >>> shape, classes = (224, 224, 3), 1000
        >>> x = keras.layers.Input(shape)
        >>> blocks = [2, 2, 2, 2]
        >>> block = keras_resnet.blocks.basic_1d
        >>> model = keras_resnet.models.ResNet(x, classes, blocks, block, classes=classes)
        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks, block, include_top=True, classes=1000, freeze_bn=True, numerical_names=None,
                 *args, **kwargs):
        if keras.backend.image_data_format() == "channels_last":
            axis = 3
        else:
            axis = 1

        if numerical_names is None:
            numerical_names = [True] * len(blocks)

        x = layers.ZeroPadding1D(padding=3, name="padding_conv1")(inputs)
        x = layers.Conv1D(64, (7, 7), strides=(2, 2), use_bias=False, name="conv1")(x)
        x = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn, name="bn_conv1")(x)
        x = layers.Activation("relu", name="conv1_relu")(x)
        x = layers.MaxPooling1D((3, 3), strides=(2, 2), padding="same", name="pool1")(x)

        features = 64

        outputs = []

        for stage_id, iterations in enumerate(blocks):
            for block_id in range(iterations):
                x = block(
                    features,
                    stage_id,
                    block_id,
                    numerical_name=(block_id > 0 and numerical_names[stage_id]),
                    freeze_bn=freeze_bn
                )(x)

            features *= 2

            outputs.append(x)

        if include_top:
            assert classes > 0

            x = keras.layers.GlobalAveragePooling1D(name="pool5")(x)
            x = keras.layers.Dense(classes, activation="softmax", name="fc1000")(x)

            super(ResNet1D, self).__init__(inputs=inputs, outputs=x, *args, **kwargs)
        else:
            # Else output each stages features
            super(ResNet1D, self).__init__(inputs=inputs, outputs=outputs, *args, **kwargs)


class ResNet1D18(ResNet1D):
    """
    Constructs a `keras.models.Model` according to the ResNet18 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet18(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [2, 2, 2, 2]

        super(ResNet1D18, self).__init__(inputs, blocks, block=B.basic_1d, include_top=include_top, classes=classes,
                                         freeze_bn=freeze_bn, *args, **kwargs)


class ResNet1D34(ResNet1D):
    """
    Constructs a `keras.models.Model` according to the ResNet34 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet34(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 4, 6, 3]

        super(ResNet1D34, self).__init__(inputs, blocks, block=B.basic_1d, include_top=include_top, classes=classes,
                                         freeze_bn=freeze_bn, *args, **kwargs)


class ResNet1D50(ResNet1D):
    """
    Constructs a `keras.models.Model` according to the ResNet50 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet50(x)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 4, 6, 3]

        numerical_names = [False, False, False, False]

        super(ResNet1D50, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_1d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet1D101(ResNet1D):
    """
    Constructs a `keras.models.Model` according to the ResNet101 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet101(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 4, 23, 3]

        numerical_names = [False, True, True, False]

        super(ResNet1D101, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_1d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet1D152(ResNet1D):
    """
    Constructs a `keras.models.Model` according to the ResNet152 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet152(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 8, 36, 3]

        numerical_names = [False, True, True, False]

        super(ResNet1D152, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_1d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet1D200(ResNet1D):
    """
    Constructs a `keras.models.Model` according to the ResNet200 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet200(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 24, 36, 3]

        numerical_names = [False, True, True, False]

        super(ResNet1D200, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_1d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet2D(Model):
    """
    Constructs a `keras.models.Model` object using the given block count.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param block: a residual block (e.g. an instance of `keras_resnet.blocks.basic_2d`)

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :param numerical_names: list of bool, same size as blocks, used to indicate whether names of layers should include numbers or letters

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.blocks
        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> blocks = [2, 2, 2, 2]

        >>> block = keras_resnet.blocks.basic_2d

        >>> model = keras_resnet.models.ResNet(x, classes, blocks, block, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks, block, include_top=True, classes=1000, freeze_bn=True, numerical_names=None,
                 *args, **kwargs):
        if keras.backend.image_data_format() == "channels_last":
            axis = 3
        else:
            axis = 1

        if numerical_names is None:
            numerical_names = [True] * len(blocks)

        x = Conv2D(64, (7, 7), strides=(2, 2), use_bias=False, name="conv1", padding="same")(inputs)
        x = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn, name="bn_conv1")(x)
        x = Activation("relu", name="conv1_relu")(x)
        x = MaxPooling2D((3, 3), strides=(2, 2), padding="same", name="pool1")(x)

        features = 64

        outputs = []

        for stage_id, iterations in enumerate(blocks):
            for block_id in range(iterations):
                x = block(features, stage_id, block_id, numerical_name=(block_id > 0 and numerical_names[stage_id]),
                          freeze_bn=freeze_bn)(x)

            features *= 2
            outputs.append(x)

        if include_top:
            assert classes > 0

            x = keras.layers.GlobalAveragePooling2D(name="pool5")(x)
            x = keras.layers.Dense(classes, activation="softmax", name="fc1000")(x)

            super(ResNet2D, self).__init__(inputs=inputs, outputs=x, *args, **kwargs)
        else:
            # Else output each stages features
            super(ResNet2D, self).__init__(inputs=inputs, outputs=outputs, *args, **kwargs)


class ResNet2D18(ResNet2D):
    """
    Constructs a `keras.models.Model` according to the ResNet18 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet18(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [2, 2, 2, 2]

        super(ResNet2D18, self).__init__(
            inputs,
            blocks,
            block=B.basic_2d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet2D34(ResNet2D):
    """
    Constructs a `keras.models.Model` according to the ResNet34 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet34(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 4, 6, 3]

        super(ResNet2D34, self).__init__(
            inputs,
            blocks,
            block=B.basic_2d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet2D50(ResNet2D):
    """
    Constructs a `keras.models.Model` according to the ResNet50 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet50(x)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 4, 6, 3]

        numerical_names = [False, False, False, False]

        super(ResNet2D50, self).__init__(inputs, blocks,
                                         numerical_names=numerical_names,
                                         block=B.bottleneck_2d,
                                         include_top=include_top,
                                         classes=classes,
                                         freeze_bn=freeze_bn,
                                         *args,
                                         **kwargs)


class ResNet2D101(ResNet2D):
    """
    Constructs a `keras.models.Model` according to the ResNet101 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet101(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 4, 23, 3]

        numerical_names = [False, True, True, False]

        super(ResNet2D101, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_2d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet2D152(ResNet2D):
    """
    Constructs a `keras.models.Model` according to the ResNet152 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet152(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 8, 36, 3]

        numerical_names = [False, True, True, False]

        super(ResNet2D152, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_2d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet2D200(ResNet2D):
    """
    Constructs a `keras.models.Model` according to the ResNet200 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet200(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 24, 36, 3]

        numerical_names = [False, True, True, False]

        super(ResNet2D200, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_2d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet3D(Model):
    """
    Constructs a `keras.models.Model` object using the given block count.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param block: a residual block (e.g. an instance of `keras_resnet.blocks.basic_3d`)

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :param numerical_names: list of bool, same size as blocks, used to indicate whether names of layers should include numbers or letters

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.blocks
        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> blocks = [2, 2, 2, 2]

        >>> block = keras_resnet.blocks.basic_3d

        >>> model = keras_resnet.models.ResNet(x, classes, blocks, block, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(
            self,
            inputs,
            blocks,
            block,
            include_top=True,
            classes=1000,
            freeze_bn=True,
            numerical_names=None,
            *args,
            **kwargs
    ):
        if keras.backend.image_data_format() == "channels_last":
            axis = 3
        else:
            axis = 1

        if numerical_names is None:
            numerical_names = [True] * len(blocks)

        x = keras.layers.ZeroPadding3D(padding=3, name="padding_conv1")(inputs)
        x = keras.layers.Conv3D(64, (7, 7), strides=(2, 2), use_bias=False, name="conv1")(x)
        x = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn, name="bn_conv1")(x)
        x = keras.layers.Activation("relu", name="conv1_relu")(x)
        x = keras.layers.MaxPooling3D((3, 3), strides=(2, 2), padding="same", name="pool1")(x)

        features = 64

        outputs = []

        for stage_id, iterations in enumerate(blocks):
            for block_id in range(iterations):
                x = block(
                    features,
                    stage_id,
                    block_id,
                    numerical_name=(block_id > 0 and numerical_names[stage_id]),
                    freeze_bn=freeze_bn
                )(x)

            features *= 2

            outputs.append(x)

        if include_top:
            assert classes > 0

            x = keras.layers.GlobalAveragePooling3D(name="pool5")(x)
            x = keras.layers.Dense(classes, activation="softmax", name="fc1000")(x)

            super(ResNet3D, self).__init__(inputs=inputs, outputs=x, *args, **kwargs)
        else:
            # Else output each stages features
            super(ResNet3D, self).__init__(inputs=inputs, outputs=outputs, *args, **kwargs)


class ResNet3D18(ResNet3D):
    """
    Constructs a `keras.models.Model` according to the ResNet18 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet18(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [2, 2, 2, 2]

        super(ResNet3D18, self).__init__(
            inputs,
            blocks,
            block=B.basic_3d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet3D34(ResNet3D):
    """
    Constructs a `keras.models.Model` according to the ResNet34 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet34(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 4, 6, 3]

        super(ResNet3D34, self).__init__(
            inputs,
            blocks,
            block=B.basic_3d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet3D50(ResNet3D):
    """
    Constructs a `keras.models.Model` according to the ResNet50 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet50(x)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 4, 6, 3]

        numerical_names = [False, False, False, False]

        super(ResNet3D50, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_3d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet3D101(ResNet3D):
    """
    Constructs a `keras.models.Model` according to the ResNet101 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet101(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 4, 23, 3]

        numerical_names = [False, True, True, False]

        super(ResNet3D101, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_3d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet3D152(ResNet3D):
    """
    Constructs a `keras.models.Model` according to the ResNet152 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet152(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 8, 36, 3]

        numerical_names = [False, True, True, False]

        super(ResNet3D152, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_3d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class ResNet3D200(ResNet3D):
    """
    Constructs a `keras.models.Model` according to the ResNet200 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> model = keras_resnet.models.ResNet200(x, classes=classes)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """

    def __init__(self, inputs, blocks=None, include_top=True, classes=1000, freeze_bn=False, *args, **kwargs):
        if blocks is None:
            blocks = [3, 24, 36, 3]

        numerical_names = [False, True, True, False]

        super(ResNet3D200, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_3d,
            include_top=include_top,
            classes=classes,
            freeze_bn=freeze_bn,
            *args,
            **kwargs
        )


class FPN2D(Model):
    def __init__(
            self,
            inputs,
            blocks,
            block,
            freeze_bn=True,
            numerical_names=None,
            *args,
            **kwargs
    ):
        if keras.backend.image_data_format() == "channels_last":
            axis = 3
        else:
            axis = 1

        if numerical_names is None:
            numerical_names = [True] * len(blocks)

        x = keras.layers.Conv2D(64, (7, 7), strides=(2, 2), use_bias=False, name="conv1", padding="same")(inputs)
        x = BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn, name="bn_conv1")(x)
        x = keras.layers.Activation("relu", name="conv1_relu")(x)
        x = keras.layers.MaxPooling2D((3, 3), strides=(2, 2), padding="same", name="pool1")(x)

        features = 64

        outputs = []

        for stage_id, iterations in enumerate(blocks):
            for block_id in range(iterations):
                x = block(
                    features,
                    stage_id,
                    block_id,
                    numerical_name=(block_id > 0 and numerical_names[stage_id]),
                    freeze_bn=freeze_bn
                )(x)

            features *= 2

            outputs.append(x)

        c2, c3, c4, c5 = outputs

        pyramid_5 = keras.layers.Conv2D(
            filters=256,
            kernel_size=1,
            strides=1,
            padding="same",
            name="c5_reduced"
        )(c5)

        upsampled_p5 = keras.layers.UpSampling2D(
            interpolation="bilinear",
            name="p5_upsampled",
            size=(2, 2)
        )(pyramid_5)

        pyramid_4 = keras.layers.Conv2D(
            filters=256,
            kernel_size=1,
            strides=1,
            padding="same",
            name="c4_reduced"
        )(c4)

        pyramid_4 = keras.layers.Add(
            name="p4_merged"
        )([upsampled_p5, pyramid_4])

        upsampled_p4 = keras.layers.UpSampling2D(
            interpolation="bilinear",
            name="p4_upsampled",
            size=(2, 2)
        )(pyramid_4)

        pyramid_4 = keras.layers.Conv2D(
            filters=256,
            kernel_size=3,
            strides=1,
            padding="same",
            name="p4"
        )(pyramid_4)

        pyramid_3 = keras.layers.Conv2D(
            filters=256,
            kernel_size=1,
            strides=1,
            padding="same",
            name="c3_reduced"
        )(c3)

        pyramid_3 = keras.layers.Add(
            name="p3_merged"
        )([upsampled_p4, pyramid_3])

        upsampled_p3 = keras.layers.UpSampling2D(
            interpolation="bilinear",
            name="p3_upsampled",
            size=(2, 2)
        )(pyramid_3)

        pyramid_3 = keras.layers.Conv2D(
            filters=256,
            kernel_size=3,
            strides=1,
            padding="same",
            name="p3"
        )(pyramid_3)

        pyramid_2 = keras.layers.Conv2D(
            filters=256,
            kernel_size=1,
            strides=1,
            padding="same",
            name="c2_reduced"
        )(c2)

        pyramid_2 = keras.layers.Add(
            name="p2_merged"
        )([upsampled_p3, pyramid_2])

        pyramid_2 = keras.layers.Conv2D(
            filters=256,
            kernel_size=3,
            strides=1,
            padding="same",
            name="p2"
        )(pyramid_2)

        pyramid_6 = keras.layers.MaxPooling2D(strides=2, name="p6")(pyramid_5)

        outputs = [
            pyramid_2,
            pyramid_3,
            pyramid_4,
            pyramid_5,
            pyramid_6
        ]

        super(FPN2D, self).__init__(
            inputs=inputs,
            outputs=outputs,
            *args,
            **kwargs
        )


class FPN2D50(FPN2D):
    def __init__(self, inputs, blocks=None, *args, **kwargs):
        if blocks is None:
            blocks = [3, 4, 6, 3]

        numerical_names = [False, False, False, False]

        super(FPN2D50, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_2d,
            *args,
            **kwargs
        )


class FPN2D18(FPN2D):
    def __init__(self, inputs, blocks=None, *args, **kwargs):
        if blocks is None:
            blocks = [2, 2, 2, 2]

        super(FPN2D18, self).__init__(
            inputs,
            blocks,
            block=B.basic_2d,
            *args,
            **kwargs
        )


class FPN2D34(FPN2D):
    def __init__(self, inputs, blocks=None, *args, **kwargs):
        if blocks is None:
            blocks = [3, 4, 6, 3]

        super(FPN2D34, self).__init__(
            inputs,
            blocks,
            block=B.basic_2d,
            *args,
            **kwargs
        )


class FPN2D101(FPN2D):
    def __init__(self, inputs, blocks=None, *args, **kwargs):
        if blocks is None:
            blocks = [3, 4, 23, 3]

        numerical_names = [False, True, True, False]

        super(FPN2D101, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_2d,
            *args,
            **kwargs
        )


class FPN2D152(FPN2D):
    def __init__(self, inputs, blocks=None, *args, **kwargs):
        if blocks is None:
            blocks = [3, 8, 36, 3]

        numerical_names = [False, True, True, False]

        super(FPN2D152, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_2d,
            *args,
            **kwargs
        )


class FPN2D200(FPN2D):
    def __init__(self, inputs, blocks=None, *args, **kwargs):
        if blocks is None:
            blocks = [3, 24, 36, 3]

        numerical_names = [False, True, True, False]

        super(FPN2D200, self).__init__(
            inputs,
            blocks,
            numerical_names=numerical_names,
            block=B.bottleneck_2d,
            *args,
            **kwargs
        )


def TimeDistributedResNet(inputs, blocks, block, include_top=True, classes=1000, freeze_bn=True, *args, **kwargs):
    """
    Constructs a time distributed `keras.models.Model` object using the given block count.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param block: a time distributed residual block (e.g. an instance of `keras_resnet.blocks.time_distributed_basic_2d`)

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :param freeze_bn: if true, freezes BatchNormalization layers (ie. no updates are done in these layers)

    :return model: Time distributed ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.blocks
        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> blocks = [2, 2, 2, 2]

        >>> blocks = keras_resnet.blocks.time_distributed_basic_2d

        >>> y = keras_resnet.models.TimeDistributedResNet(x, classes, blocks, blocks)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """
    if keras.backend.image_data_format() == "channels_last":
        axis = 3
    else:
        axis = 1

    x = keras.layers.TimeDistributed(keras.layers.ZeroPadding2D(padding=3), name="padding_conv1")(inputs)
    x = keras.layers.TimeDistributed(keras.layers.Conv2D(64, (7, 7), strides=(2, 2), use_bias=False), name="conv1")(x)
    x = keras.layers.TimeDistributed(BatchNormalizationFreeze(axis=axis, epsilon=1e-5, freeze=freeze_bn),
                                     name="bn_conv1")(x)
    x = keras.layers.TimeDistributed(keras.layers.Activation("relu"), name="conv1_relu")(x)
    x = keras.layers.TimeDistributed(keras.layers.MaxPooling2D((3, 3), strides=(2, 2), padding="same"), name="pool1")(x)

    features = 64

    outputs = []

    for stage_id, iterations in enumerate(blocks):
        for block_id in range(iterations):
            x = block(features, stage_id, block_id, numerical_name=(blocks[stage_id] > 6), freeze_bn=freeze_bn)(x)

        features *= 2
        outputs.append(x)

    if include_top:
        assert classes > 0

        x = keras.layers.TimeDistributed(keras.layers.GlobalAveragePooling2D(), name="pool5")(x)
        x = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"), name="fc1000")(x)

        return keras.models.Model(inputs=inputs, outputs=x, *args, **kwargs)
    else:
        # Else output each stages features
        return keras.models.Model(inputs=inputs, outputs=outputs, *args, **kwargs)


def TimeDistributedResNet18(inputs, blocks=None, include_top=True, classes=1000, *args, **kwargs):
    """
    Constructs a time distributed `keras.models.Model` according to the ResNet18 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :return model: Time distributed ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.TimeDistributedResNet18(x)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """
    if blocks is None:
        blocks = [2, 2, 2, 2]

    return TimeDistributedResNet(inputs, blocks, block=B.time_distributed_basic_2d,
                                 include_top=include_top, classes=classes, *args, **kwargs)


def TimeDistributedResNet34(inputs, blocks=None, include_top=True, classes=1000, *args, **kwargs):
    """
    Constructs a time distributed `keras.models.Model` according to the ResNet34 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :return model: Time distributed ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.TimeDistributedResNet34(x)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """
    if blocks is None:
        blocks = [3, 4, 6, 3]

    return TimeDistributedResNet(inputs, blocks, block=B.time_distributed_basic_2d,
                                 include_top=include_top, classes=classes, *args, **kwargs)


def TimeDistributedResNet50(inputs, blocks=None, include_top=True, classes=1000, *args, **kwargs):
    """
    Constructs a time distributed `keras.models.Model` according to the ResNet50 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.TimeDistributedResNet50(x)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """
    if blocks is None:
        blocks = [3, 4, 6, 3]

    return TimeDistributedResNet(inputs, blocks, block=B.time_distributed_bottleneck_2d,
                                 include_top=include_top, classes=classes, *args, **kwargs)


def TimeDistributedResNet101(inputs, blocks=None, include_top=True, classes=1000, *args, **kwargs):
    """
    Constructs a time distributed `keras.models.Model` according to the ResNet101 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :return model: Time distributed ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.TimeDistributedResNet101(x)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """
    if blocks is None:
        blocks = [3, 4, 23, 3]

    return TimeDistributedResNet(inputs, blocks, block=B.time_distributed_bottleneck_2d,
                                 include_top=include_top, classes=classes, *args, **kwargs)


def TimeDistributedResNet152(inputs, blocks=None, include_top=True, classes=1000, *args, **kwargs):
    """
    Constructs a time distributed `keras.models.Model` according to the ResNet152 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :return model: Time distributed ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.TimeDistributedResNet152(x)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """
    if blocks is None:
        blocks = [3, 8, 36, 3]

    return TimeDistributedResNet(inputs, blocks, block=B.time_distributed_bottleneck_2d,
                                 include_top=include_top, classes=classes, *args, **kwargs)


def TimeDistributedResNet200(inputs, blocks=None, include_top=True, classes=1000, *args, **kwargs):
    """
    Constructs a time distributed `keras.models.Model` according to the ResNet200 specifications.

    :param inputs: input tensor (e.g. an instance of `keras.layers.Input`)

    :param blocks: the network’s residual architecture

    :param include_top: if true, includes classification layers

    :param classes: number of classes to classify (include_top must be true)

    :return model: Time distributed ResNet model with encoding output (if `include_top=False`) or classification output (if `include_top=True`)

    Usage:

        >>> import keras_resnet.models

        >>> shape, classes = (224, 224, 3), 1000

        >>> x = keras.layers.Input(shape)

        >>> y = keras_resnet.models.TimeDistributedResNet200(x)

        >>> y = keras.layers.TimeDistributed(keras.layers.Flatten())(y.output)

        >>> y = keras.layers.TimeDistributed(keras.layers.Dense(classes, activation="softmax"))(y)

        >>> model = keras.models.Model(x, y)

        >>> model.compile("adam", "categorical_crossentropy", ["accuracy"])
    """
    if blocks is None:
        blocks = [3, 24, 36, 3]

    return TimeDistributedResNet(inputs, blocks, block=B.time_distributed_bottleneck_2d,
                                 include_top=include_top, classes=classes, *args, **kwargs)



from .core import ACGAN
from .core import BGAN
from .core import BIGAN
from .core import BaseGAN
from .core import CCGAN
from .core import CGAN
from .core import COGAN
from .core import DCGAN
from .core import DUALGAN
from .core import GAN
from .core import INFOGAN
from .core import LSGAN
from .core import SGAN
from .core import WGAN



import math

import numpy as np
from tensorflow.keras.initializers import Initializer


class PriorProbability(Initializer):
    """ Apply a prior probability to the weights.
    """

    def __init__(self, probability=0.01):
        self.probability = probability

    def get_config(self):
        return {
            'probability': self.probability
        }

    def __call__(self, shape, dtype=None):
        # set bias to -log((1 - p)/p) for foreground
        if type(dtype) is not str:
            dtype = 'float32'
        result = np.ones(shape, dtype=dtype) * - \
            math.log((1 - self.probability) / self.probability)

        return result
